"""Functionality for data scaling."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Callable, Dict, Tuple, Type

import pandas as pd
import torch
from torch import Tensor

from baybe.utils.dataframe import to_tensor

_ScaleFun = Callable[[Tensor], Tensor]


class Scaler(ABC):
    """Abstract base class for all scalers.

    Args:
        searchspace: The search space that should be scaled.
    """

    type: str
    """Class variable encoding the type of the scaler."""

    SUBCLASSES: Dict[str, Type[Scaler]] = {}
    """Class variable for all subclasses"""

    def __init__(self, searchspace: pd.DataFrame):
        self.searchspace = searchspace
        self.fitted = False
        self.scale_x: _ScaleFun
        self.scale_y: _ScaleFun
        self.unscale_x: _ScaleFun
        self.unscale_y: _ScaleFun
        self.unscale_m: _ScaleFun
        self.unscale_s: _ScaleFun

    @abstractmethod
    def fit_transform(self, x: Tensor, y: Tensor) -> Tuple[Tensor, Tensor]:
        """Fit the scaler using the given training data and transform the data.

        Args:
            x: The x-data that should be used.
            y: The y-data that should be used.

        Returns:
            The transformed data.
        """

    def transform(self, x: Tensor) -> Tensor:
        """Scale a given input.

        Args:
            x: The given input.

        Returns:
            The scaled input.

        Raises:
            RuntimeError: If the scaler is not fitted first.
        """
        if not self.fitted:
            raise RuntimeError("Scaler object must be fitted first.")
        return self.scale_x(x)

    def untransform(self, mean: Tensor, variance: Tensor) -> Tuple[Tensor, Tensor]:
        """Transform mean values and variances back to the original domain.

        Args:
            mean: The given mean values.
            variance: The given variances.

        Returns:
            The "un-transformed" means and variances.

        Raises:
            RuntimeError: If the scaler object is not fitted first.
        """
        if not self.fitted:
            raise RuntimeError("Scaler object must be fitted first.")
        return self.unscale_m(mean), self.unscale_s(variance)

    @classmethod
    def __init_subclass__(cls, **kwargs):
        """Register new subclasses dynamically."""
        super().__init_subclass__(**kwargs)
        cls.SUBCLASSES[cls.type] = cls


class DefaultScaler(Scaler):
    """A scaler that normalizes inputs to the unit cube and standardizes targets."""

    type = "DEFAULT"
    # See base class.

    def fit_transform(  # noqa: D102
        self, x: Tensor, y: Tensor
    ) -> Tuple[Tensor, Tensor]:
        # See base class.

        # Get the searchspace boundaries
        searchspace = to_tensor(self.searchspace)
        bounds = torch.vstack(
            [torch.min(searchspace, dim=0)[0], torch.max(searchspace, dim=0)[0]]
        )

        # Compute the mean and standard deviation of the training targets
        mean = torch.mean(y, dim=0)
        std = torch.std(y, dim=0)

        # Functions for input and target scaling
        self.scale_x = lambda x: (x - bounds[0]) / (bounds[1] - bounds[0])
        self.scale_y = lambda x: (x - mean) / std

        # Functions for inverse input and target scaling
        self.unscale_x = lambda x: x * (bounds[1] - bounds[0]) + bounds[0]
        self.unscale_y = lambda x: x * std + mean

        # Functions for inverse mean and variance scaling
        self.unscale_m = lambda x: x * std + mean
        self.unscale_s = lambda x: x * std**2

        # Flag that the scaler has been fitted
        self.fitted = True

        return self.scale_x(x), self.scale_y(y)



"""Test serialization of meta recommenders."""

import pytest

from baybe.recommenders import (
    FPSRecommender,
    RandomRecommender,
    SequentialMetaRecommender,
    StreamingSequentialMetaRecommender,
    TwoPhaseMetaRecommender,
)
from baybe.recommenders.meta.base import MetaRecommender
from tests.conftest import select_recommender

# Create some recommenders of different class for better differentiation after roundtrip
RECOMMENDERS = [RandomRecommender(), FPSRecommender()]
assert len(RECOMMENDERS) == len(set(rec.__class__.__name__ for rec in RECOMMENDERS))


def roundtrip(recommender: MetaRecommender) -> MetaRecommender:
    """Roundtrip serialization."""
    string = recommender.to_json()
    return MetaRecommender.from_json(string)


@pytest.mark.parametrize(
    "recommender",
    [
        TwoPhaseMetaRecommender(),
        SequentialMetaRecommender(recommenders=[RandomRecommender()]),
    ],
)
def test_meta_recommender_serialization(recommender):
    """Roundtrip serialization of meta recommenders."""
    assert recommender == roundtrip(recommender)


def test_unsupported_serialization():
    """Attempting to serialize an unserializable recommender should raise an error."""
    recommender = StreamingSequentialMetaRecommender(
        recommenders=(rec for rec in [RandomRecommender()])
    )
    with pytest.raises(NotImplementedError):
        recommender.to_json()


@pytest.mark.parametrize(
    "recommender",
    [
        TwoPhaseMetaRecommender(
            initial_recommender=RECOMMENDERS[0],
            recommender=RECOMMENDERS[1],
            switch_after=1,
        ),
        SequentialMetaRecommender(recommenders=RECOMMENDERS),
    ],
)
def test_meta_recommender_state_serialization(recommender):
    """Roundtrip-serialized meta recommenders keep their internal states."""
    # Before serialization, identity must hold
    rec = select_recommender(recommender, 0)
    assert rec is RECOMMENDERS[0]
    rec = select_recommender(recommender, 1)
    assert rec is RECOMMENDERS[1]

    # After serialization, identity no longer holds but equality does
    recommender = roundtrip(recommender)
    rec = select_recommender(recommender, 1)
    assert rec == RECOMMENDERS[1]



"""Parameter utilities."""

from typing import Any, Callable, Collection, Dict, List, Optional, TypeVar

import pandas as pd

from baybe.parameters.base import Parameter

_TParameter = TypeVar("_TParameter", bound=Parameter)


def get_parameters_from_dataframe(
    df: pd.DataFrame,
    factory: Callable[[str, Collection[Any]], _TParameter],
    parameters: Optional[List[_TParameter]] = None,
) -> List[_TParameter]:
    """Create a list of parameters from a dataframe.

    Returns one parameter for each column of the given dataframe. By default,
    the parameters are created using the provided factory, which takes the name
    of the column and its unique values as arguments. However, there is also
    the possibility to provide explicit parameter objects with names matching specific
    columns of the dataframe, to bypass the parameter factory creation for those
    columns. This allows finer control, for example, to specify custom parameter
    attributes (e.g. specific optional arguments) compared to what would be provided
    by the factory. Still, the pre-specified parameters are validated to ensure that
    they are compatible with the contents of the dataframe.

    Args:
        df: The dataframe from which to create the parameters.
        factory: A parameter factor, creating parameter objects for the columns
            from the column name and the unique column values.
        parameters: An optional list of parameter objects to bypass the factory
            creation for columns whose names match with the parameter names.

    Returns:
        The combined parameter list, containing both the (validated) pre-specified
        parameters and the parameters inferred from the dataframe.

    Raises:
        ValueError: If several parameters with identical names are provided.
        ValueError: If a parameter was specified for which no match was found.
    """
    # Turn the pre-specified parameters into a dict and check for duplicate names
    specified_params: Dict[str, _TParameter] = {}
    if parameters is not None:
        for param in parameters:
            if param.name in specified_params:
                raise ValueError(
                    f"You provided several parameters with the name '{param.name}'."
                )
            specified_params[param.name] = param

    # Try to find a parameter match for each dataframe column
    parameters = []
    for name, series in df.items():
        assert isinstance(
            name, str
        ), "The given dataframe must only contain string-valued column names."
        unique_values = series.unique()

        # If a match is found, assert that the values are in range
        if match := specified_params.pop(name, None):
            if not all(match.is_in_range(x) for x in unique_values):
                raise ValueError(
                    f"The dataframe column '{name}' contains the values "
                    f"{unique_values}, which are outside the range of {match}."
                )
            parameters.append(match)

        # Otherwise, create a new parameter using the factory
        else:
            param = factory(name, unique_values)
            parameters.append(param)

    # By now, all pre-specified parameters must have been used
    if specified_params:
        raise ValueError(
            f"For the parameter(s) {list(specified_params.keys())}, "
            f"no match could be found in the given dataframe."
        )

    return parameters



"""Tests for the substance parameter."""

import pytest

from baybe.parameters.enum import SubstanceEncoding

from .conftest import _CHEM_INSTALLED, run_iterations


@pytest.mark.skipif(
    not _CHEM_INSTALLED, reason="Optional chem dependency not installed."
)
@pytest.mark.parametrize(
    "parameter_names",
    [["Categorical_1", f"Substance_1_{enc}"] for enc in SubstanceEncoding],
    ids=[enc.name for enc in SubstanceEncoding],
)
def test_run_iterations(campaign, batch_size, n_iterations):
    """Test running some iterations with fake results and a substance parameter."""
    run_iterations(campaign, n_iterations, batch_size)



"""This file contains some utility function for building the documentation."""

import pathlib
import shutil
import textwrap
from subprocess import DEVNULL, STDOUT, check_call

from tqdm import tqdm


def adjust_banner(file_path: str, light_banner: str, dark_banner: str) -> None:
    """Adjust the banner to have different versions for light and dark furo style.

    Args:
        file_path: The (relative) path to the file that needs to be adjusted. Typically
            the index and README_link file.
        light_banner: The name of the light mode banner.
        dark_banner: The name of the dark mode banner.
    """
    with open(file_path, "r") as file:
        lines = file.readlines()

    line_index = None
    for i, line in enumerate(lines):
        if "banner" in line:
            line_index = i
            break

    if line_index is not None:
        line = lines[line_index]
        light_line = line.replace("reference external", "reference external only-light")
        lines[line_index] = light_line
        # lines.insert(line_index + 1, light_line)
        dark_line = light_line.replace("only-light", "only-dark")
        dark_line = dark_line.replace(light_banner, dark_banner)
        lines[line_index + 1] = dark_line

        with open(file_path, "w") as file:
            file.writelines(lines)


def create_example_documentation(example_dest_dir: str, ignore_examples: bool):
    """Create the documentation version of the examples files.

    Note that this deletes the destination directory if it already exists.

    Args:
        example_dest_dir: The destination directory.
        ignore_examples: A flag denoting whether or not to ignore the examples. If set,
            then the examples are constructed, but not executed.
    """
    # Folder where the .md files created are stored
    examples_directory = pathlib.Path(example_dest_dir)

    # if the destination directory already exists it is deleted
    if examples_directory.is_dir():
        shutil.rmtree(examples_directory)

    # Copy the examples folder in the destination directory
    shutil.copytree("examples", examples_directory)

    # For the toctree of the top level example folder, we need to keep track of all
    # folders. We thus write the header here and populate it during the execution of the
    # examples
    ex_file = """# Examples\n\n```{toctree}\n:maxdepth: 2\n\n"""

    # List all directories in the examples folder
    ex_directories = [d for d in examples_directory.iterdir() if d.is_dir()]

    # This list contains the order of the examples as we want to have them in the end.
    # The examples that should be the first ones are already included here and skipped
    # later on. ALl other are just included.
    ex_order = [
        "Basics<Basics/Basics>\n",
        "Searchspaces<Searchspaces/Searchspaces>\n",
        "Constraints Discrete<Constraints_Discrete/Constraints_Discrete>\n",
        "Constraints Continuous<Constraints_Continuous/Constraints_Continuous>\n",
        "Multi Target<Multi_Target/Multi_Target>\n",
        "Serialization<Serialization/Serialization>\n",
        "Custom Surrogates<Custom_Surrogates/Custom_Surrogates>\n",
    ]

    # Iterate over the directories.
    for sub_directory in (pbar := tqdm(ex_directories)):
        # Get the name of the current folder
        # Format it by replacing underscores and capitalizing the words
        folder_name = sub_directory.stem
        formatted = " ".join(word.capitalize() for word in folder_name.split("_"))

        # Create the link to the folder to the top level toctree.
        ex_file_entry = formatted + f"<{folder_name}/{folder_name}>\n"
        # Add it to the list of examples if it is not already contained
        if ex_file_entry not in ex_order:
            ex_order.append(ex_file_entry)

        # We need to create a file for the inclusion of the folder.
        # We thus get the content of the corresponding header file.
        header_folder_name = sub_directory / f"{folder_name}_Header.md"
        header = header_folder_name.read_text()

        subdir_toctree = header + "\n```{toctree}\n:maxdepth: 1\n\n"

        # Set description of progressbar
        pbar.set_description("Overall progress")

        # list all .py files in the subdirectory that need to be converted
        py_files = list(sub_directory.glob("**/*.py"))

        # Iterate through the individual example files
        for file in (inner_pbar := tqdm(py_files, leave=False)):
            # Include the name of the file to the toctree
            # Format it by replacing underscores and capitalizing the words
            file_name = file.stem

            formatted = " ".join(word.capitalize() for word in file_name.split("_"))
            # Remove duplicate "constraints" for the files in the constraints folder.
            if "Constraints" in folder_name and "Constraints" in formatted:
                formatted = formatted.replace("Constraints", "")

            # Also format the Prodsum name to Product/Sum
            if "Prodsum" in formatted:
                formatted = formatted.replace("Prodsum", "Product/Sum")
            subdir_toctree += formatted + f"<{file_name}>\n"

            # Set description for progress bar
            inner_pbar.set_description(f"Progressing {folder_name}")

            # Create the Markdown file:

            # 1. Convert the file to jupyter notebook
            check_call(
                ["jupytext", "--to", "notebook", file], stdout=DEVNULL, stderr=STDOUT
            )

            notebook_path = file.with_suffix(".ipynb")

            # 2. Execute the notebook and convert to markdown.
            # This is only done if we decide not to ignore the examples.
            # The creation of the files themselves and converting them to markdown still
            # happens since we need the files to check for link integrity.
            convert_execute = [
                "jupyter",
                "nbconvert",
                "--to",
                "notebook",
                "--inplace",
                notebook_path,
            ]
            if not ignore_examples:
                convert_execute.append("--execute")

            to_markdown = ["jupyter", "nbconvert", "--to", "markdown", notebook_path]

            check_call(convert_execute, stdout=DEVNULL, stderr=STDOUT)
            check_call(
                to_markdown,
                stdout=DEVNULL,
                stderr=STDOUT,
            )

            # CLEANUP
            markdown_path = file.with_suffix(".md")
            # We wrap lines which are too long as long as they do not contain a link.
            # To discover whether a line contains a link, we check if the string "]("
            # is contained.
            with open(markdown_path, "r", encoding="UTF-8") as markdown_file:
                content = markdown_file.read()
                wrapped_lines = []
                ignored_substrings = (
                    "![svg]",
                    "![png]",
                    "<Figure size",
                    "it/s",
                    "s/it",
                )
                for line in content.splitlines():
                    if any(substring in line for substring in ignored_substrings):
                        continue
                    if len(line) > 88 and "](" not in line:
                        wrapped = textwrap.wrap(line, width=88)
                        wrapped_lines.extend(wrapped)
                    else:
                        wrapped_lines.append(line)

            # Add a manual new line to each of the lines
            lines = [line + "\n" for line in wrapped_lines]
            # Delete lines we do not want to have in our documentation
            # lines = [line for line in lines if "![svg]" not in line]
            # We check whether pre-built light and dark plots exist. If so, we append
            # corresponding lines to our markdown file for including them.
            light_figure = pathlib.Path(sub_directory / (file_name + "_light.svg"))
            dark_figure = pathlib.Path(sub_directory / (file_name + "_dark.svg"))
            if light_figure.is_file() and dark_figure.is_file():
                lines.append(f"```{{image}} {file_name}_light.svg\n")
                lines.append(":align: center\n")
                lines.append(":class: only-light\n")
                lines.append("```\n")
                lines.append(f"```{{image}} {file_name}_dark.svg\n")
                lines.append(":align: center\n")
                lines.append(":class: only-dark\n")
                lines.append("```\n")

            # Rewrite the file
            with open(markdown_path, "w", encoding="UTF-8") as markdown_file:
                markdown_file.writelines(lines)

        # Write last line of toctree file for this directory and write the file
        subdir_toctree += "```"
        with open(
            sub_directory / f"{sub_directory.name}.md", "w", encoding="UTF-8"
        ) as f:
            f.write(subdir_toctree)

    # Append the ordered list of examples to the file for the top level folder
    ex_file += "".join(ex_order)
    # Write last line of top level toctree file and write the file
    ex_file += "```"
    with open(
        examples_directory / f"{examples_directory.name}.md", "w", encoding="UTF-8"
    ) as f:
        f.write(ex_file)

    # Remove remaining files and subdirectories from the destination directory
    # Remove any not markdown files
    for file in examples_directory.glob("**/*"):
        if file.is_file():
            if file.suffix not in (".md", ".svg") or "Header" in file.name:
                file.unlink(file)

    # Remove any remaining empty subdirectories
    for subdirectory in examples_directory.glob("*/*"):
        if subdirectory.is_dir() and not any(subdirectory.iterdir()):
            subdirectory.rmdir()



"""Utilities for numeric operations."""
from typing import List

import numpy as np
import torch

DTypeFloatNumpy = np.float64
"""Floating point data type used for numpy arrays."""

DTypeFloatTorch = torch.float64
"""Floating point data type used for torch tensors."""

DTypeFloatONNX = np.float32
"""Floating point data type used for ONNX models.

Currently, ONNX runtime does not seem to have full support for double precision.
There is no clear documentation but some references can be found here (version 1.16.0):

 * https://onnx.ai/sklearn-onnx/auto_tutorial/plot_abegin_convert_pipeline.html#converts-the-model
 * https://onnx.ai/sklearn-onnx/auto_tutorial/plot_ebegin_float_double.html
"""  # noqa: E501


def geom_mean(arr: np.ndarray, weights: List[float]) -> np.ndarray:
    """Calculate the (weighted) geometric mean along the second axis of a 2-D array.

    Alternative to ``gmean`` from scipy that avoids logarithms and division errors.

    Args:
        arr: The array containing the values for the mean computation.
        weights: Weights for the mean computation.

    Returns:
        A 1-D array containing the row-wise geometric means of the given array.
    """
    return np.prod(np.power(arr, np.atleast_2d(weights) / np.sum(weights)), axis=1)


def closest_element(array: np.ndarray, target: float) -> float:
    """Find the element of an array that is closest to a target value.

    Args:
        array: The array in which the closest value should be found.
        target: The target value.

    Returns:
        The closes element.
    """
    return array[np.abs(array - target).argmin()]


def closer_element(x: float, y: float, target: float) -> float:
    """Determine which of two given inputs is closer to a target value.

    Args:
        x: The first input that should be checked.
        y: The second input that should be checked.
        target: The target value.

    Returns:
        The closer of the two elements.
    """
    return x if np.abs(x - target) < np.abs(y - target) else y



## Example for full simulation loop using the multi target mode for custom analytic functions

# This example shows how to use a multi target objective for a custom analytic function.
# It uses a desirability value to handle several targets.

# This example assumes basic familiarity with BayBE, custom test functions and multiple targets.
# For further details, we thus refer to
# - [`campaign`](./../Basics/campaign.md) for a more general and basic example,
# - [`custom_analytical`](./custom_analytical.md) for custom test functions, and
# - [`desirability`](./../Multi_Target/desirability.md) for multiple targets.

### Necessary imports for this example

import os
from typing import Tuple

import numpy as np

from baybe import Campaign
from baybe.objective import Objective
from baybe.parameters import NumericalDiscreteParameter
from baybe.searchspace import SearchSpace
from baybe.simulation import simulate_scenarios
from baybe.targets import NumericalTarget

### Parameters for a full simulation loop

# For the full simulation, we need to define some additional parameters.
# These are the number of Monte Carlo runs and the number of experiments to be conducted per run.

SMOKE_TEST = "SMOKE_TEST" in os.environ

N_MC_ITERATIONS = 2 if SMOKE_TEST else 5
N_DOE_ITERATIONS = 2 if SMOKE_TEST else 4
BATCH_SIZE = 1 if SMOKE_TEST else 2
DIMENSION = 4
BOUNDS = [(-2, 2), (-2, 2), (-2, 2), (-2, 2)]
POINTS_PER_DIM = 3 if SMOKE_TEST else 10


### Defining the test function

# See [`custom_analytical`](./custom_analytical.md) for details.


def sum_of_squares(*x: float) -> Tuple[float, float]:
    """Calculate the sum of squares."""
    res = 0
    for y in x:
        res += y**2
    return res, 2 * res**2 - 1


### Creating the searchspace

# In this example, we construct a purely discrete space with 10 points per dimension.

parameters = [
    NumericalDiscreteParameter(
        name=f"x_{k+1}",
        values=list(np.linspace(*BOUNDS[k], POINTS_PER_DIM)),
        tolerance=0.01,
    )
    for k in range(DIMENSION)
]

searchspace = SearchSpace.from_product(parameters=parameters)


### Creating multiple target object

# The multi target mode is handled when creating the objective object.
# Thus we first need to define the different targets.
# We use two targets here.
# The first target is maximized and the second target is minimized during the optimization process.

Target_1 = NumericalTarget(
    name="Target_1", mode="MAX", bounds=(0, 100), transformation="LINEAR"
)
Target_2 = NumericalTarget(
    name="Target_2", mode="MIN", bounds=(0, 100), transformation="LINEAR"
)


### Creating the objective object

# We collect the two targets in a list and use this list to construct the objective.

targets = [Target_1, Target_2]

objective = Objective(
    mode="DESIRABILITY",
    targets=targets,
    weights=[20, 30],
    combine_func="MEAN",
)


### Constructing a campaign and performing the simulation loop

campaign = Campaign(searchspace=searchspace, objective=objective)

# We can now use the `simulate_scenarios` function to simulate a full experiment.

scenarios = {"BayBE": campaign}

results = simulate_scenarios(
    scenarios,
    sum_of_squares,
    batch_size=BATCH_SIZE,
    n_doe_iterations=N_DOE_ITERATIONS,
    n_mc_iterations=N_MC_ITERATIONS,
)

print(results)



"""Tests for the objective module."""

import pytest

from baybe.objective import Objective
from baybe.targets import NumericalTarget


class TestInvalidObjectiveCreation:
    """Invalid objective creation raises expected error."""

    # Two example targets used in the tests
    two_targets = [
        NumericalTarget(
            name="Target_1",
            mode="MAX",
            bounds=(0, 100),
        ),
        NumericalTarget(
            name="Target_2",
            mode="MIN",
            bounds=(0, 100),
        ),
    ]

    def test_empty_target_list(self):
        with pytest.raises(ValueError):
            Objective(
                mode="SINGLE",
                targets=[],
            )

    def test_wrong_target_type(self):
        with pytest.raises(ValueError):
            Objective(
                mode="SINGLE",
                targets={"A": 1, "B": 2},
            )

    def test_missing_bounds_for_desirability(self):
        with pytest.raises(ValueError):
            Objective(
                mode="DESIRABILITY",
                targets=[
                    NumericalTarget(
                        name="Target_1",
                        mode="MAX",
                        bounds=(0, 100),
                    ),
                    NumericalTarget(
                        name="Target_2",
                        mode="MIN",
                    ),
                ],
            )

    def test_invalid_combination_function(self):
        with pytest.raises(ValueError):
            Objective(
                mode="DESIRABILITY",
                combine_func="FALSE_STUFF",
                targets=self.two_targets,
            )

    def test_wrong_number_of_weights(self):
        with pytest.raises(ValueError):
            Objective(
                mode="DESIRABILITY",
                weights=[1, 2, 3],
                targets=self.two_targets,
            )

    def test_non_numeric_weights(self):
        with pytest.raises(TypeError):
            Objective(
                mode="DESIRABILITY",
                weights=[1, "ABC"],
                targets=self.two_targets,
            )

    def test_wrong_weights_type(self):
        with pytest.raises(TypeError):
            Objective(
                mode="DESIRABILITY",
                weights="ABC",
                targets=self.two_targets,
            )



"""Base classes for all pure recommenders."""

from abc import ABC
from typing import ClassVar, Optional

import pandas as pd
from attrs import define, field

from baybe.exceptions import NotEnoughPointsLeftError
from baybe.recommenders.base import RecommenderProtocol
from baybe.searchspace import SearchSpace
from baybe.searchspace.continuous import SubspaceContinuous
from baybe.searchspace.core import SearchSpaceType
from baybe.searchspace.discrete import SubspaceDiscrete


@define
class PureRecommender(ABC, RecommenderProtocol):
    """Abstract base class for all pure recommenders."""

    # Class variables
    compatibility: ClassVar[SearchSpaceType]
    """Class variable reflecting the search space compatibility."""

    # Object variables
    allow_repeated_recommendations: bool = field(default=False, kw_only=True)
    """Allow to make recommendations that were already recommended earlier.
    This only has an influence in discrete search spaces."""

    allow_recommending_already_measured: bool = field(default=True, kw_only=True)
    """Allow to make recommendations that were measured previously.
    This only has an influence in discrete search spaces."""

    def recommend(  # noqa: D102
        self,
        searchspace: SearchSpace,
        batch_size: int = 1,
        train_x: Optional[pd.DataFrame] = None,
        train_y: Optional[pd.DataFrame] = None,
    ) -> pd.DataFrame:
        # See base class
        if searchspace.type is SearchSpaceType.CONTINUOUS:
            return self._recommend_continuous(
                subspace_continuous=searchspace.continuous, batch_size=batch_size
            )
        else:
            return self._recommend_with_discrete_parts(searchspace, batch_size)

    def _recommend_discrete(
        self,
        subspace_discrete: SubspaceDiscrete,
        candidates_comp: pd.DataFrame,
        batch_size: int,
    ) -> pd.Index:
        """Generate recommendations from a discrete search space.

        Args:
            subspace_discrete: The discrete subspace from which to generate
                recommendations.
            candidates_comp: The computational representation of all discrete candidate
                points to be considered.
            batch_size: The size of the recommendation batch.

        Raises:
            NotImplementedError: If the function is not implemented by the child class.

        Returns:
            The dataframe indices of the recommended points in the provided
            computational representation.
        """
        # If this method is not implemented by a child class, try to resort to hybrid
        # recommendation (with an empty subspace) instead.
        try:
            return self._recommend_hybrid(
                searchspace=SearchSpace(discrete=subspace_discrete),
                candidates_comp=candidates_comp,
                batch_size=batch_size,
            ).index
        except NotImplementedError as exc:
            raise NotImplementedError(
                """Hybrid recommendation could not be used as fallback when trying to
                optimize a discrete space.
                This is probably due to your search space and recommender not being
                compatible.
                If you operate in discrete search spaces,
                ensure that you either use a discrete or a hybrid recommender."""
            ) from exc

    def _recommend_continuous(
        self,
        subspace_continuous: SubspaceContinuous,
        batch_size: int,
    ) -> pd.DataFrame:
        """Generate recommendations from a continuous search space.

        Args:
            subspace_continuous: The continuous subspace from which to generate
                recommendations.
            batch_size: The size of the recommendation batch.

        Raises:
            NotImplementedError: If the function is not implemented by the child class.

        Returns:
            A dataframe containing the recommendations as individual rows.
        """
        # If this method is not implemented by a child class, try to resort to hybrid
        # recommendation (with an empty subspace) instead.
        try:
            return self._recommend_hybrid(
                searchspace=SearchSpace(continuous=subspace_continuous),
                candidates_comp=pd.DataFrame(),
                batch_size=batch_size,
            )
        except NotImplementedError as exc:
            raise NotImplementedError(
                """Hybrid recommendation could not be used as fallback when trying to
                optimize a continuous space.
                This is probably due to your search space and recommender not being
                compatible.
                If you operate in continuous search spaces,
                ensure that you either use a continuous or a hybrid recommender."""
            ) from exc

    def _recommend_hybrid(
        self,
        searchspace: SearchSpace,
        candidates_comp: pd.DataFrame,
        batch_size: int,
    ) -> pd.DataFrame:
        """Generate recommendations from a hybrid search space.

        If the recommender does not implement additional functions for discrete and
        continuous search spaces, this method is used as a fallback for those spaces
        as well.

        Args:
            searchspace: The hybrid search space from which to generate
                recommendations.
            candidates_comp: The computational representation of all discrete candidate
                points to be considered.
            batch_size: The size of the recommendation batch.

        Raises:
            NotImplementedError: If the function is not implemented by the child class.

        Returns:
            A dataframe containing the recommendations as individual rows.
        """
        raise NotImplementedError("Hybrid recommendation is not implemented.")

    def _recommend_with_discrete_parts(
        self,
        searchspace: SearchSpace,
        batch_size: int,
    ) -> pd.DataFrame:
        """Obtain recommendations in search spaces with a discrete part.

        Convenience helper which sequentially performs the following tasks: get discrete
        candidates, generate recommendations, update metadata.

        Args:
            searchspace: The search space from which to generate recommendations.
            batch_size: The size of the recommendation batch.

        Returns:
            A dataframe containing the recommendations as individual rows.

        Raises:
            NotEnoughPointsLeftError: If there are fewer points left for potential
                recommendation than requested.
        """
        is_hybrid_space = searchspace.type is SearchSpaceType.HYBRID

        # Get discrete candidates
        # Repeated recommendations are always allowed for hybrid spaces
        _, candidates_comp = searchspace.discrete.get_candidates(
            allow_repeated_recommendations=is_hybrid_space
            or self.allow_repeated_recommendations,
            allow_recommending_already_measured=is_hybrid_space
            or self.allow_recommending_already_measured,
        )

        # Check if enough candidates are left
        # TODO [15917]: This check is not perfectly correct.
        if (not is_hybrid_space) and (len(candidates_comp) < batch_size):
            raise NotEnoughPointsLeftError(
                f"Using the current settings, there are fewer than {batch_size} "
                "possible data points left to recommend. This can be "
                "either because all data points have been measured at some point "
                "(while 'allow_repeated_recommendations' or "
                "'allow_recommending_already_measured' being False) "
                "or because all data points are marked as 'dont_recommend'."
            )

        # Get recommendations
        if is_hybrid_space:
            rec = self._recommend_hybrid(searchspace, candidates_comp, batch_size)
            idxs = rec.index
        else:
            idxs = self._recommend_discrete(
                searchspace.discrete, candidates_comp, batch_size
            )
            rec = searchspace.discrete.exp_rep.loc[idxs, :]

        # Update metadata
        searchspace.discrete.metadata.loc[idxs, "was_recommended"] = True

        # Return recommendations
        return rec



"""Meta recommenders that switch recommenders based on the experimentation progress."""
# TODO After bayesian recommenders are enabled with no training data, a refactoring of
#  this file will resolve type errors
# mypy: disable-error-code="arg-type"

from typing import Iterable, Iterator, List, Literal, Optional

import pandas as pd
from attrs import define, field
from attrs.validators import deep_iterable, in_, instance_of

from baybe.exceptions import NoRecommendersLeftError
from baybe.recommenders.meta.base import MetaRecommender
from baybe.recommenders.pure.base import PureRecommender
from baybe.recommenders.pure.bayesian.sequential_greedy import (
    SequentialGreedyRecommender,
)
from baybe.recommenders.pure.nonpredictive.base import NonPredictiveRecommender
from baybe.recommenders.pure.nonpredictive.sampling import RandomRecommender
from baybe.searchspace import SearchSpace
from baybe.serialization import (
    block_deserialization_hook,
    block_serialization_hook,
    converter,
)

# TODO: Make bayesian recommenders handle empty training data
_unsupported_recommender_error = ValueError(
    f"For cases where no training is available, the selected recommender "
    f"must be a subclass of '{NonPredictiveRecommender.__name__}'."
)


@define
class TwoPhaseMetaRecommender(MetaRecommender):
    """A two-phased meta recommender that switches at a certain specified point.

    The recommender is switched when a new (batch) recommendation is requested and
    the training data set size (i.e., the total number of collected measurements
    including those gathered before the meta recommender was active) is equal to or
    greater than the number specified via the ``switch_after`` parameter.

    Note:
        Throughout each phase, the meta recommender reuses the **same** recommender
        object, that is, no new instances are created. Therefore, special attention is
        required when using the meta recommender with stateful recommenders.
    """

    initial_recommender: PureRecommender = field(factory=RandomRecommender)
    """The initial recommender used by the meta recommender."""

    recommender: PureRecommender = field(factory=SequentialGreedyRecommender)
    """The recommender used by the meta recommender after the switch."""

    switch_after: int = field(default=1)
    """The number of experiments after which the recommender is switched for the next
    requested batch."""

    def select_recommender(  # noqa: D102
        self,
        searchspace: SearchSpace,
        batch_size: int = 1,
        train_x: Optional[pd.DataFrame] = None,
        train_y: Optional[pd.DataFrame] = None,
    ) -> PureRecommender:
        # See base class.

        # TODO: enable bayesian recommenders for empty training data
        if (train_x is None or len(train_x) == 0) and not isinstance(
            self.initial_recommender, NonPredictiveRecommender
        ):
            raise _unsupported_recommender_error

        return (
            self.recommender
            if len(train_x) >= self.switch_after
            else self.initial_recommender
        )


@define
class SequentialMetaRecommender(MetaRecommender):
    """A meta recommender that uses a pre-defined sequence of recommenders.

    A new recommender is taken from the sequence whenever at least one new measurement
    is available, until all recommenders are exhausted. More precisely, a recommender
    change is triggered whenever the size of the training dataset increases; the
    actual content of the dataset is ignored.

    Note:
        The provided sequence of recommenders will be internally pre-collected into a
        list. If this is not acceptable, consider using
        :class:`baybe.recommenders.meta.sequential.StreamingSequentialMetaRecommender`
        instead.

    Raises:
        NoRecommendersLeftError: If more recommenders are requested than there are
            recommenders available and ``mode="raise"``.
    """

    # Exposed
    recommenders: List[PureRecommender] = field(
        converter=list, validator=deep_iterable(instance_of(PureRecommender))
    )
    """A finite-length sequence of recommenders to be used. For infinite-length
    iterables, see
    :class:`baybe.recommenders.meta.sequential.StreamingSequentialMetaRecommender`."""

    mode: Literal["raise", "reuse_last", "cyclic"] = field(
        default="raise",
        validator=in_(("raise", "reuse_last", "cyclic")),
    )
    """Defines what shall happen when the last recommender in the sequence has been
    consumed but additional recommender changes are triggered:

        * ``"raise"``: An error is raised.
        * ``"reuse_last"``: The last recommender in the sequence is used indefinitely.
        * ``"cycle"``: The selection restarts from the beginning of the sequence.
    """

    # Private
    # TODO: These should **not** be exposed via the constructor but the workaround
    #   is currently needed for correct (de-)serialization. A proper approach would be
    #   to not set them via the constructor but through a custom hook in combination
    #   with `_cattrs_include_init_false=True`. However, the way
    #   `get_base_structure_hook` is currently designed prevents such a hook from
    #   taking action.
    _step: int = field(default=-1, alias="_step")
    """Counts how often the recommender has already been switched."""

    _n_last_measurements: int = field(default=-1, alias="_n_last_measurements")
    """The number of measurements that were available at the last call."""

    def select_recommender(  # noqa: D102
        self,
        searchspace: SearchSpace,
        batch_size: int = 1,
        train_x: Optional[pd.DataFrame] = None,
        train_y: Optional[pd.DataFrame] = None,
    ) -> PureRecommender:
        # See base class.

        # If the training dataset size has increased, move to the next recommender
        if len(train_x) > self._n_last_measurements:
            self._step += 1
        # If the training dataset size has decreased, something went wrong
        elif len(train_x) < self._n_last_measurements:
            raise RuntimeError(
                f"The training dataset size decreased from {self._n_last_measurements} "
                f"to {len(train_x)} since the last function call, which indicates that "
                f"'{self.__class__.__name__}' was not used as intended."
            )

        # Get the right index for the "next" recommender
        idx = self._step
        if self.mode == "reuse_last":
            idx = min(idx, len(self.recommenders) - 1)
        elif self.mode == "cyclic":
            idx %= len(self.recommenders)

        # Get the recommender
        try:
            recommender = self.recommenders[idx]
        except IndexError as ex:
            raise NoRecommendersLeftError(
                f"A total of {self._step+1} recommender(s) was/were requested but the "
                f"provided sequence contains only {self._step} element(s)."
            ) from ex

        # Remember the training dataset size for the next call
        self._n_last_measurements = len(train_x)

        # TODO: enable bayesian recommenders for empty training data
        if (train_x is None or len(train_x) == 0) and not isinstance(
            recommender, NonPredictiveRecommender
        ):
            raise _unsupported_recommender_error

        return recommender


@define
class StreamingSequentialMetaRecommender(MetaRecommender):
    """A meta recommender that switches between recommenders from an iterable.

    Similar to :class:`baybe.recommenders.meta.sequential.SequentialMetaRecommender`
    but without explicit list conversion. Consequently, it supports arbitrary
    iterables, possibly of infinite length. The downside is that serialization is not
    supported.

    Raises:
        NoRecommendersLeftError: If more recommenders are requested than there are
            recommenders available.
    """

    # Exposed
    recommenders: Iterable[PureRecommender] = field()
    """An iterable providing the recommenders to be used."""

    # Private
    # TODO: See :class:`baybe.recommenders.meta.sequential.SequentialMetaRecommender`
    _step: int = field(init=False, default=-1)
    """Counts how often the recommender has already been switched."""

    _n_last_measurements: int = field(init=False, default=-1)
    """The number of measurements that were available at the last call."""

    _iterator: Iterator = field(init=False)
    """The iterator used to traverse the recommenders."""

    _last_recommender: Optional[PureRecommender] = field(init=False, default=None)
    """The recommender returned from the last call."""

    @_iterator.default
    def default_iterator(self):
        """Initialize the recommender iterator."""
        return iter(self.recommenders)

    def select_recommender(  # noqa: D102
        self,
        searchspace: SearchSpace,
        batch_size: int = 1,
        train_x: Optional[pd.DataFrame] = None,
        train_y: Optional[pd.DataFrame] = None,
    ) -> PureRecommender:
        # See base class.

        use_last = True

        # If the training dataset size has increased, move to the next recommender
        if len(train_x) > self._n_last_measurements:
            self._step += 1
            use_last = False
        # If the training dataset size has decreased, something went wrong
        elif len(train_x) < self._n_last_measurements:
            raise RuntimeError(
                f"The training dataset size decreased from {self._n_last_measurements} "
                f"to {len(train_x)} since the last function call, which indicates that "
                f"'{self.__class__.__name__}' was not used as intended."
            )

        # Get the recommender
        try:
            if not use_last:
                self._last_recommender = next(self._iterator)
        except StopIteration as ex:
            raise NoRecommendersLeftError(
                f"A total of {self._step+1} recommender(s) was/were requested but the "
                f"provided iterator provided only {self._step} element(s)."
            ) from ex

        # Remember the training dataset size for the next call
        self._n_last_measurements = len(train_x)

        # TODO: enable bayesian recommenders for empty training data
        if (train_x is None or len(train_x) == 0) and not isinstance(
            self._last_recommender, NonPredictiveRecommender
        ):
            raise _unsupported_recommender_error

        return self._last_recommender  # type: ignore[return-value]


# The recommender iterable cannot be serialized
converter.register_unstructure_hook(
    StreamingSequentialMetaRecommender, block_serialization_hook
)
converter.register_structure_hook(
    StreamingSequentialMetaRecommender, block_deserialization_hook
)



"""Base classes for all constraints."""

from abc import ABC, abstractmethod
from typing import Any, ClassVar, List, Tuple

import pandas as pd
import torch
from attr import define, field
from attr.validators import min_len
from torch import Tensor

from baybe.constraints.conditions import Condition
from baybe.parameters import NumericalContinuousParameter
from baybe.serialization import (
    SerialMixin,
    converter,
    get_base_structure_hook,
    unstructure_base,
)
from baybe.utils.numerical import DTypeFloatTorch


@define
class Constraint(ABC, SerialMixin):
    """Abstract base class for all constraints.

    Constraints use conditions and chain them together to filter unwanted entries from
    the search space.
    """

    # class variables
    # TODO: it might turn out these are not needed at a later development stage
    eval_during_creation: ClassVar[bool]
    """Class variable encoding whether the condition is evaluated during creation."""

    eval_during_modeling: ClassVar[bool]
    """Class variable encoding whether the condition is evaluated during modeling."""

    # Object variables
    parameters: List[str] = field(validator=min_len(1))
    """The list of parameters used for the constraint."""

    @parameters.validator
    def _validate_params(  # noqa: DOC101, DOC103
        self, _: Any, params: List[str]
    ) -> None:
        """Validate the parameter list.

        Raises:
            ValueError: If ``params`` contains duplicate values.
        """
        if len(params) != len(set(params)):
            raise ValueError(
                f"The given 'parameters' list must have unique values "
                f"but was: {params}."
            )

    def summary(self) -> dict:
        """Return a custom summarization of the constraint."""
        constr_dict = dict(
            Type=self.__class__.__name__, Affected_Parameters=self.parameters
        )
        return constr_dict

    @property
    def is_continuous(self) -> bool:
        """Boolean indicating if this is a constraint over continuous parameters."""
        return isinstance(self, ContinuousConstraint)

    @property
    def is_discrete(self) -> bool:
        """Boolean indicating if this is a constraint over discrete parameters."""
        return isinstance(self, DiscreteConstraint)


@define
class DiscreteConstraint(Constraint, ABC):
    """Abstract base class for discrete constraints.

    Discrete constraints use conditions and chain them together to filter unwanted
    entries from the search space.
    """

    # class variables
    eval_during_creation: ClassVar[bool] = True
    # See base class.

    eval_during_modeling: ClassVar[bool] = False
    # See base class.

    @abstractmethod
    def get_invalid(self, data: pd.DataFrame) -> pd.Index:
        """Get the indices of dataframe entries that are invalid under the constraint.

        Args:
            data: A dataframe where each row represents a particular parameter
                combination.

        Returns:
            The dataframe indices of rows where the constraint is violated.
        """


@define
class ContinuousConstraint(Constraint, ABC):
    """Abstract base class for continuous constraints.

    Continuous constraints use parameter lists and coefficients to define in-/equality
    constraints over a continuous parameter space.
    """

    # class variables
    eval_during_creation: ClassVar[bool] = False
    # See base class.

    eval_during_modeling: ClassVar[bool] = True
    # See base class.

    # object variables
    coefficients: List[float] = field()
    """In-/equality coefficient for each entry in ``parameters``."""

    rhs: float = field(default=0.0)
    """Right-hand side value of the in-/equality."""

    @coefficients.validator
    def _validate_coefficients(  # noqa: DOC101, DOC103
        self, _: Any, coefficients: List[float]
    ) -> None:
        """Validate the coefficients.

        Raises:
            ValueError: If the number of coefficients does not match the number of
                parameters.
        """
        if len(self.parameters) != len(coefficients):
            raise ValueError(
                "The given 'coefficients' list must have one floating point entry for "
                "each entry in 'parameters'."
            )

    @coefficients.default
    def _default_coefficients(self):
        """Return equal weight coefficients as default."""
        return [1.0] * len(self.parameters)

    def to_botorch(
        self, parameters: List[NumericalContinuousParameter], idx_offset: int = 0
    ) -> Tuple[Tensor, Tensor, float]:
        """Cast the constraint in a format required by botorch.

        Used in calling ``optimize_acqf_*`` functions, for details see
        https://botorch.org/api/optim.html#botorch.optim.optimize.optimize_acqf

        Args:
            parameters: The parameter objects of the continuous space.
            idx_offset: Offset to the provided parameter indices.

        Returns:
            The tuple required by botorch.
        """
        param_names = [p.name for p in parameters]
        param_indices = [
            param_names.index(p) + idx_offset
            for p in self.parameters
            if p in param_names
        ]

        return (
            torch.tensor(param_indices),
            torch.tensor(self.coefficients, dtype=DTypeFloatTorch),
            self.rhs,
        )


# Register (un-)structure hooks
converter.register_unstructure_hook(Condition, unstructure_base)
converter.register_structure_hook(Condition, get_base_structure_hook(Condition))
converter.register_unstructure_hook(Constraint, unstructure_base)
converter.register_structure_hook(Constraint, get_base_structure_hook(Constraint))



"""Validation functionality for parameters."""

from typing import Any, Sequence

import numpy as np
from attrs.validators import gt, instance_of, lt


def validate_unique_values(  # noqa: DOC101, DOC103
    obj: Any, _: Any, value: list
) -> None:
    """Validate that there are no duplicates in ``value``.

    Raises:
        ValueError: If there are duplicates in ``value``.
    """
    if len(set(value)) != len(value):
        raise ValueError(
            f"Cannot assign the following values containing duplicates to "
            f"parameter {obj.name}: {value}."
        )


def validate_decorrelation(obj: Any, attribute: Any, value: float) -> None:
    """Validate the decorrelation."""
    instance_of((bool, float))(obj, attribute, value)
    if isinstance(value, float):
        gt(0.0)(obj, attribute, value)
        lt(1.0)(obj, attribute, value)


def validate_is_finite(  # noqa: DOC101, DOC103
    obj: Any, _: Any, value: Sequence[float]
) -> None:
    """Validate that ``value`` contains no infinity/nan.

    Raises:
        ValueError: If ``value`` contains infinity/nan.
    """
    if not all(np.isfinite(value)):
        raise ValueError(
            f"Cannot assign the following values containing infinity/nan to "
            f"parameter {obj.name}: {value}."
        )



"""Deprecation tests."""

import json

import pytest

from baybe import BayBE, Campaign
from baybe.exceptions import DeprecationError
from baybe.recommenders.meta.sequential import TwoPhaseMetaRecommender
from baybe.recommenders.pure.nonpredictive.sampling import (
    FPSRecommender,
    RandomRecommender,
)
from baybe.searchspace import SearchSpace
from baybe.strategies import (
    SequentialStrategy,
    Strategy,
    StreamingSequentialStrategy,
    TwoPhaseStrategy,
)
from baybe.targets import Objective
from baybe.targets.base import Target
from baybe.utils.interval import Interval


def test_deprecated_baybe_class(parameters, objective):
    """Using the deprecated ``BayBE`` class raises a warning."""
    with pytest.warns(DeprecationWarning):
        BayBE(SearchSpace.from_product(parameters), objective)


def test_moved_objective(targets):
    """Importing ``Objective`` from ``baybe.targets`` raises a warning."""
    with pytest.warns(DeprecationWarning):
        Objective(mode="SINGLE", targets=targets)


def test_renamed_surrogate():
    """Importing from ``baybe.surrogate`` raises a warning."""
    with pytest.warns(DeprecationWarning):
        from baybe.surrogate import GaussianProcessSurrogate  # noqa: F401


def test_missing_recommender_type(config):
    """Specifying a recommender without a corresponding type raises a warning."""
    dict_ = json.loads(config)
    dict_["recommender"].pop("type")
    config_without_strategy_type = json.dumps(dict_)
    with pytest.warns(DeprecationWarning):
        Campaign.from_config(config_without_strategy_type)


# Create some recommenders of different class for better differentiation after roundtrip
RECOMMENDERS = [RandomRecommender(), FPSRecommender()]
assert len(RECOMMENDERS) == len(set(rec.__class__.__name__ for rec in RECOMMENDERS))


@pytest.mark.parametrize(
    "test_objects",
    [
        (Strategy, {}),
        (TwoPhaseStrategy, {}),
        (SequentialStrategy, {"recommenders": RECOMMENDERS}),
        (StreamingSequentialStrategy, {"recommenders": RECOMMENDERS}),
    ],
)
def test_deprecated_strategies(test_objects):
    """Using the deprecated strategy classes raises a warning."""
    strategy, arguments = test_objects
    with pytest.warns(DeprecationWarning):
        strategy(**arguments)


def test_deprecated_interval_is_finite():
    """Using the deprecated ``Interval.is_finite`` property raises a warning."""
    with pytest.warns(DeprecationWarning):
        Interval(0, 1).is_finite


def test_missing_target_type():
    """Specifying a target without a corresponding type raises a warning."""
    with pytest.warns(DeprecationWarning):
        Target.from_json(
            json.dumps(
                {
                    "name": "missing_type",
                    "mode": "MAX",
                }
            )
        )


deprecated_config = """
{
    "parameters": [
        {
            "type": "CategoricalParameter",
            "name": "Granularity",
            "values": ["coarse", "fine", "ultra-fine"]
        }
    ],
    "objective": {
        "mode": "SINGLE",
        "targets": [
            {
                "name": "Yield",
                "mode": "MAX"
            }
        ]
    }
}
"""


def test_deprecated_config():
    """Using the deprecated config format raises a warning."""
    with pytest.warns(UserWarning):
        Campaign.from_config(deprecated_config)


@pytest.mark.parametrize("flag", [False, True])
def test_deprecated_campaign_tolerance_flag(flag):
    """Constructing a Campaign with the deprecated tolerance flag raises an error."""
    with pytest.raises(DeprecationError):
        Campaign(None, None, None, numerical_measurements_must_be_within_tolerance=flag)


def test_deprecated_batch_quantity_keyword(campaign):
    """Using the deprecated batch_quantity keyword raises an error."""
    with pytest.raises(DeprecationError):
        campaign.recommend(batch_quantity=5)


@pytest.mark.parametrize("flag", (True, False))
def test_deprecated_strategy_allow_flags(flag):
    """Using the deprecated recommender "allow" flags raises an error."""
    with pytest.raises(DeprecationError):
        TwoPhaseMetaRecommender(allow_recommending_already_measured=flag)
    with pytest.raises(DeprecationError):
        TwoPhaseMetaRecommender(allow_repeated_recommendations=flag)


def test_deprecated_strategy_campaign_flag(recommender):
    """Using the deprecated strategy keyword raises an error."""
    with pytest.raises(DeprecationError):
        Campaign(None, None, None, strategy=recommender)



"""Test serialization of parameters."""

import hypothesis.strategies as st
import pytest
from hypothesis import given
from pytest import param

from baybe.parameters.base import Parameter

from ..conftest import _CHEM_INSTALLED
from ..hypothesis_strategies.parameters import (
    categorical_parameter,
    custom_parameter,
    numerical_continuous_parameter,
    numerical_discrete_parameter,
    substance_parameter,
    task_parameter,
)


@pytest.mark.parametrize(
    "parameter_strategy",
    [
        param(numerical_discrete_parameter(), id="NumericalDiscreteParameter"),
        param(numerical_continuous_parameter(), id="NumericalContinuousParameter"),
        param(categorical_parameter(), id="CategoricalParameter"),
        param(task_parameter(), id="TaskParameter"),
        param(custom_parameter(), id="CustomParameter"),
        param(
            substance_parameter(),
            id="SubstanceParameter",
            marks=pytest.mark.skipif(
                not _CHEM_INSTALLED, reason="Optional chem dependency not installed."
            ),
        ),
    ],
)
@given(data=st.data())
def test_parameter_roundtrip(parameter_strategy, data):
    """A serialization roundtrip yields an equivalent object."""
    parameter = data.draw(parameter_strategy)
    string = parameter.to_json()
    parameter2 = Parameter.from_json(string)
    assert parameter == parameter2, (parameter, parameter2)



"""Tests for meta recommenders."""

import pytest

from baybe.exceptions import NoRecommendersLeftError
from baybe.recommenders import (
    FPSRecommender,
    RandomRecommender,
    SequentialGreedyRecommender,
    SequentialMetaRecommender,
    StreamingSequentialMetaRecommender,
    TwoPhaseMetaRecommender,
)
from tests.conftest import select_recommender

RECOMMENDERS = [RandomRecommender(), FPSRecommender(), SequentialGreedyRecommender()]


def test_twophase_meta_recommender():
    """The recommender switches the recommender at the requested point."""
    initial_recommender = RandomRecommender()
    subsequent_recommender = RandomRecommender()
    switch_after = 3
    recommender = TwoPhaseMetaRecommender(
        initial_recommender=initial_recommender,
        recommender=subsequent_recommender,
        switch_after=switch_after,
    )
    for training_size in range(6):
        rec = select_recommender(recommender, training_size)
        target = (
            initial_recommender
            if training_size < switch_after
            else subsequent_recommender
        )
        assert rec is target


@pytest.mark.parametrize("mode", ["raise", "reuse_last", "cyclic"])
@pytest.mark.parametrize("recommenders", [RECOMMENDERS])
def test_sequential_meta_recommender(recommenders, mode):
    """The recommender provides its recommenders in the right order."""
    meta_recommender = SequentialMetaRecommender(recommenders=recommenders, mode=mode)
    training_size = 0

    # First iteration over provided recommender sequence
    for reference in recommenders:
        training_size += 1

        # The returned recommender coincides with what was put in
        recommender = select_recommender(meta_recommender, training_size)
        assert recommender is reference

        # Selection with unchanged training size yields again the same recommender
        recommender = select_recommender(meta_recommender, training_size)
        assert recommender is reference

        # Selection with smaller training size raises an error
        with pytest.raises(RuntimeError):
            select_recommender(meta_recommender, training_size - 1)

    # Second iteration over provided recommender sequence
    for cycled in recommenders:
        training_size += 1

        if mode == "raise":
            # Requesting more batches than there are recommenders raises an error
            with pytest.raises(NoRecommendersLeftError):
                select_recommender(meta_recommender, training_size)

        elif mode == "reuse_last":
            # The last recommender is selected repeatedly
            recommender = select_recommender(meta_recommender, training_size)
            assert recommender == recommenders[-1]

        elif mode == "cyclic":
            # The selection restarts from the first recommender
            recommender = select_recommender(meta_recommender, training_size)
            assert recommender == cycled


@pytest.mark.parametrize(
    "recommenders",
    [
        RECOMMENDERS,  # list
        (rec for rec in RECOMMENDERS),  # generator
    ],
)
def test_streaming_sequential_meta_recommender(recommenders):
    """The recommender provides its recommenders in the right order."""
    meta_recommender = StreamingSequentialMetaRecommender(recommenders=recommenders)
    training_size = 0

    for reference in RECOMMENDERS:
        training_size += 1

        # The returned recommender coincides with what was put in
        recommender = select_recommender(meta_recommender, training_size)
        assert recommender is reference

        # Selection with unchanged training size yields again the same recommender
        recommender = select_recommender(meta_recommender, training_size)
        assert recommender is reference

        # Selection with smaller training size raises an error
        with pytest.raises(RuntimeError):
            select_recommender(meta_recommender, training_size - 1)



"""Hypothesis strategies for targets."""

import hypothesis.strategies as st

from baybe.targets.enum import TargetMode
from baybe.targets.numerical import _VALID_TRANSFORMATIONS, NumericalTarget

from .utils import interval

target_name = st.text(min_size=1)
"""A strategy that generates target names."""


@st.composite
def numerical_target(draw: st.DrawFn):
    """Generate :class:`baybe.targets.numerical.NumericalTarget`."""
    name = draw(target_name)
    mode = draw(st.sampled_from(TargetMode))
    bounds = draw(
        interval(
            exclude_half_bounded=True, exclude_fully_unbounded=mode is TargetMode.MATCH
        )
    )
    transformation = draw(st.sampled_from(_VALID_TRANSFORMATIONS[mode]))

    return NumericalTarget(
        name=name, mode=mode, bounds=bounds, transformation=transformation
    )


target = numerical_target()
"""A strategy that generates targets."""



"""Naive recommender for hybrid spaces."""

import warnings
from typing import ClassVar, Optional, cast

import pandas as pd
from attrs import define, evolve, field, fields
from torch import Tensor

from baybe.acquisition import PartialAcquisitionFunction
from baybe.recommenders.pure.base import PureRecommender
from baybe.recommenders.pure.bayesian.base import BayesianRecommender
from baybe.recommenders.pure.bayesian.sequential_greedy import (
    SequentialGreedyRecommender,
)
from baybe.recommenders.pure.nonpredictive.base import NonPredictiveRecommender
from baybe.searchspace import SearchSpace, SearchSpaceType
from baybe.utils.dataframe import to_tensor


@define
class NaiveHybridSpaceRecommender(PureRecommender):
    """Recommend points by independent optimization of subspaces.

    This recommender splits the hybrid search space in the discrete and continuous
    subspace. Each of the subspaces is optimized on its own, and the recommenders for
    those subspaces can be chosen upon initialization. If this recommender is used on
    a non-hybrid space, it uses the corresponding recommender.
    """

    # Class variables
    compatibility: ClassVar[SearchSpaceType] = SearchSpaceType.HYBRID
    # See base class.

    # Object variables
    # TODO This used to be a Union of BayesianRecommender and NonPredictiveRecommender.
    # Due to serialization issues, this was changed to PureRecommender in general.
    # As we currently do not have other subclasses of PureRecommender, this solution
    # works for now. Still, we manually check whether the disc_recommender belongs to
    # one of these two subclasses such that we might be able to easily spot a potential
    # problem that might come up when implementing new subclasses of PureRecommender
    disc_recommender: PureRecommender = field(factory=SequentialGreedyRecommender)
    """The recommender used for the discrete subspace. Default:
    :class:`baybe.recommenders.pure.bayesian.sequential_greedy.SequentialGreedyRecommender`"""

    cont_recommender: BayesianRecommender = field(factory=SequentialGreedyRecommender)
    """The recommender used for the continuous subspace. Default:
    :class:`baybe.recommenders.pure.bayesian.sequential_greedy.SequentialGreedyRecommender`"""

    def __attrs_post_init__(self):
        """Validate if flags are synchronized and overrides them otherwise."""
        if (
            flag := self.allow_recommending_already_measured
        ) != self.disc_recommender.allow_recommending_already_measured:
            warnings.warn(
                f"The value of "
                f"'{fields(self.__class__).allow_recommending_already_measured.name}' "
                f"differs from what is specified in the discrete recommender. "
                f"The value of the discrete recommender will be ignored."
            )
            self.disc_recommender = evolve(
                self.disc_recommender,
                allow_recommending_already_measured=flag,
            )

        if (
            flag := self.allow_repeated_recommendations
        ) != self.disc_recommender.allow_repeated_recommendations:
            warnings.warn(
                f"The value of "
                f"'{fields(self.__class__).allow_repeated_recommendations.name}' "
                f"differs from what is specified in the discrete recommender. "
                f"The value of the discrete recommender will be ignored."
            )
            self.disc_recommender = evolve(
                self.disc_recommender,
                allow_repeated_recommendations=flag,
            )

    def recommend(  # noqa: D102
        self,
        searchspace: SearchSpace,
        batch_size: int = 1,
        train_x: Optional[pd.DataFrame] = None,
        train_y: Optional[pd.DataFrame] = None,
    ) -> pd.DataFrame:
        # See base class.

        if (not isinstance(self.disc_recommender, BayesianRecommender)) and (
            not isinstance(self.disc_recommender, NonPredictiveRecommender)
        ):
            raise NotImplementedError(
                """The discrete recommender should be either a Bayesian or a
                NonPredictiveRecommender."""
            )

        # Check if the space is a pure continuous or discrete space first and just use
        # the corresponding recommendation function in that case
        degenerate_recommender = None
        if searchspace.type == SearchSpaceType.DISCRETE:
            degenerate_recommender = self.disc_recommender
        elif searchspace.type == SearchSpaceType.CONTINUOUS:
            degenerate_recommender = self.cont_recommender
        if degenerate_recommender is not None:
            return degenerate_recommender.recommend(
                searchspace=searchspace,
                batch_size=batch_size,
                train_x=train_x,
                train_y=train_y,
            )

        # We are in a hybrid setting now

        # We will attach continuous parts to discrete parts and the other way round.
        # To make things simple, we sample a single point in the continuous space which
        # will then be attached to every discrete point when the acquisition function
        # is evaluated.
        cont_part = searchspace.continuous.samples_random(1)
        cont_part_tensor = cast(Tensor, to_tensor(cont_part)).unsqueeze(-2)

        # Get discrete candidates. The metadata flags are ignored since the search space
        # is hybrid
        # TODO Slight BOILERPLATE CODE, see recommender.py, ll. 47+
        _, candidates_comp = searchspace.discrete.get_candidates(
            allow_repeated_recommendations=True,
            allow_recommending_already_measured=True,
        )

        # We now check whether the discrete recommender is bayesian.
        if isinstance(self.disc_recommender, BayesianRecommender):
            # Get access to the recommenders acquisition function
            self.disc_recommender.setup_acquisition_function(
                searchspace, train_x, train_y
            )

            # Construct the partial acquisition function that attaches cont_part
            # whenever evaluating the acquisition function
            disc_acqf_part = PartialAcquisitionFunction(
                acqf=self.disc_recommender._acquisition_function,
                pinned_part=cont_part_tensor,
                pin_discrete=False,
            )

            self.disc_recommender._acquisition_function = disc_acqf_part

        # Call the private function of the discrete recommender and get the indices
        disc_rec_idx = self.disc_recommender._recommend_discrete(
            subspace_discrete=searchspace.discrete,
            candidates_comp=candidates_comp,
            batch_size=batch_size,
        )

        # Get one random discrete point that will be attached when evaluating the
        # acquisition function in the discrete space.
        disc_part = searchspace.discrete.comp_rep.loc[disc_rec_idx].sample(1)
        disc_part_tensor = cast(Tensor, to_tensor(disc_part)).unsqueeze(-2)

        # Setup a fresh acquisition function for the continuous recommender
        self.cont_recommender.setup_acquisition_function(searchspace, train_x, train_y)

        # Construct the continuous space as a standalone space
        cont_acqf_part = PartialAcquisitionFunction(
            acqf=self.cont_recommender._acquisition_function,
            pinned_part=disc_part_tensor,
            pin_discrete=True,
        )
        self.cont_recommender._acquisition_function = cont_acqf_part

        # Call the private function of the continuous recommender
        rec_cont = self.cont_recommender._recommend_continuous(
            searchspace.continuous, batch_size
        )

        # Glue the solutions together and return them
        rec_disc_exp = searchspace.discrete.exp_rep.loc[disc_rec_idx]
        rec_cont.index = rec_disc_exp.index
        rec_exp = pd.concat([rec_disc_exp, rec_cont], axis=1)
        return rec_exp



"""Custom parameters."""

from functools import cached_property
from typing import Any, ClassVar, Union

import numpy as np
import pandas as pd
from attrs import define, field
from attrs.validators import min_len

from baybe.parameters.base import DiscreteParameter
from baybe.parameters.enum import CustomEncoding
from baybe.parameters.validation import validate_decorrelation
from baybe.utils.boolean import eq_dataframe
from baybe.utils.dataframe import df_uncorrelated_features


@define(frozen=True, slots=False)
class CustomDiscreteParameter(DiscreteParameter):
    """Custom parameters.

    For these parameters, the user can read in a precomputed representation for labels,
    e.g. from quantum chemistry.
    """

    # class variables
    is_numeric: ClassVar[bool] = False
    # See base class.

    # object variables
    data: pd.DataFrame = field(validator=min_len(2), eq=eq_dataframe)
    """A mapping that provides the encoding for all available parameter values."""

    decorrelate: Union[bool, float] = field(
        default=True, validator=validate_decorrelation
    )
    """Specifies the used decorrelation mode for the parameter encoding.

        - ``False``: The encoding is used as is.
        - ``True``: The encoding is decorrelated using a default correlation threshold.
        - float in (0, 1): The encoding is decorrelated using the specified threshold.
    """

    encoding: CustomEncoding = field(init=False, default=CustomEncoding.CUSTOM)
    # See base class.

    @data.validator
    def _validate_custom_data(  # noqa: DOC101, DOC103
        self, _: Any, value: pd.DataFrame
    ) -> None:
        """Validate the dataframe with the custom representation.

        Raises:
            ValueError: If the dataframe contains non-numeric values.
            ValueError: If the dataframe index contains non-string values.
            ValueError: If the dataframe index contains empty strings.
            ValueError: If the dataframe contains ``NaN``.
            ValueError: If the dataframe index contains duplicates.
            ValueError: If the dataframe contains columns with only one unique value.
        """
        if value.select_dtypes("number").shape[1] != value.shape[1]:
            raise ValueError(
                f"The custom dataframe for parameter {self.name} contains "
                f"non-numeric values."
            )
        if not all(isinstance(x, str) for x in value.index):
            raise ValueError(
                f"The custom dataframe for parameter {self.name} contains non-string "
                f"index values."
            )
        if not all(len(x) > 0 for x in value.index):
            raise ValueError(
                f"The custom dataframe for parameter {self.name} contains empty string "
                f"index values."
            )
        if not np.isfinite(value.values).all():
            raise ValueError(
                f"The custom dataframe for parameter {self.name} contains nan/infinity "
                f"entries."
            )
        if len(value) != len(set(value.index)):
            raise ValueError(
                f"The custom dataframe for parameter {self.name} contains "
                f"duplicated indices."
            )
        if any(value.nunique() == 1):
            raise ValueError(
                f"The custom dataframe for parameter {self.name} has columns "
                "that contain only a single value and hence carry no information."
            )

    @property
    def values(self) -> tuple:
        """Returns the representing labels of the parameter."""
        return tuple(self.data.index)

    @cached_property
    def comp_df(self) -> pd.DataFrame:  # noqa: D102
        # See base class.
        # The encoding is directly provided by the user
        # We prepend the parameter name to the columns names to avoid potential
        # conflicts with other parameters
        comp_df = self.data.rename(columns=lambda x: f"{self.name}_{x}")

        # Get a decorrelated subset of the provided features
        if self.decorrelate:
            if isinstance(self.decorrelate, bool):
                comp_df = df_uncorrelated_features(comp_df)
            else:
                comp_df = df_uncorrelated_features(comp_df, threshold=self.decorrelate)

        return comp_df



"""Test if all examples can be run without error."""

import os
import runpy
from pathlib import Path

import pytest

from baybe.surrogates import _ONNX_INSTALLED

from ..conftest import _CHEM_INSTALLED

# Run these tests in reduced settings
_SMOKE_TEST_CACHE = os.environ.get("SMOKE_TEST", None)
os.environ["SMOKE_TEST"] = "true"


@pytest.mark.slow
@pytest.mark.skipif(
    not (_CHEM_INSTALLED and _ONNX_INSTALLED), reason="skipped for core tests"
)
@pytest.mark.parametrize("example", Path("examples/").rglob("*.py"))
def test_example(example: Path):
    """Test an individual example by running it."""
    runpy.run_path(str(example))


if _SMOKE_TEST_CACHE is not None:
    os.environ["SMOKE_TEST"] = _SMOKE_TEST_CACHE
else:
    os.environ.pop("SMOKE_TEST")



"""Temporary functionality for backward compatibility."""

import warnings

from baybe.recommenders.meta import (
    SequentialMetaRecommender,
    StreamingSequentialMetaRecommender,
    TwoPhaseMetaRecommender,
)


def Strategy(*args, **kwargs) -> TwoPhaseMetaRecommender:
    """A ``Strategy`` alias for backward compatibility."""  # noqa: D401 (imperative mood)
    warnings.warn(
        f"Using 'Strategy' directly is deprecated and will be removed in a future "
        f"version. Please use 'recommenders.{TwoPhaseMetaRecommender.__name__}' class "
        f"instead.",
        DeprecationWarning,
    )

    return TwoPhaseMetaRecommender(*args, **kwargs)


def TwoPhaseStrategy(*args, **kwargs) -> TwoPhaseMetaRecommender:
    """A ``TwoPhaseStrategy`` alias for backward compatibility."""  # noqa: D401 (imperative mood)
    warnings.warn(
        f"'TwoPhaseStrategy' is deprecated and will be removed in a future "
        f"version. Please use 'recommenders.{TwoPhaseMetaRecommender.__name__}' class "
        f"instead.",
        DeprecationWarning,
    )

    return TwoPhaseMetaRecommender(*args, **kwargs)


def SequentialStrategy(*args, **kwargs) -> SequentialMetaRecommender:
    """A ``SequentialStrategy`` alias for backward compatibility."""  # noqa: D401 (imperative mood)
    warnings.warn(
        f"'SequentialStrategy' is deprecated and will be removed in a future version. "
        f"Please use 'recommenders.{SequentialMetaRecommender.__name__}' class "
        f"instead.",
        DeprecationWarning,
    )

    return SequentialMetaRecommender(*args, **kwargs)


def StreamingSequentialStrategy(*args, **kwargs) -> StreamingSequentialMetaRecommender:
    """A ``StreamingSequentialStrategy`` alias for backward compatibility."""  # noqa: D401 (imperative mood)
    warnings.warn(
        f"'StreamingSequentialStrategy' is deprecated and will be removed in a future "
        f"version. Please use "
        f"'recommenders.{StreamingSequentialMetaRecommender.__name__}' class instead.",
        DeprecationWarning,
    )

    return StreamingSequentialMetaRecommender(*args, **kwargs)



"""Test serialization of campaigns."""

import pytest
from cattrs import ClassValidationError

from baybe.campaign import Campaign


def roundtrip(campaign: Campaign) -> Campaign:
    string = campaign.to_json()
    return Campaign.from_json(string)


def test_campaign_serialization(campaign):
    campaign2 = roundtrip(campaign)
    assert campaign == campaign2

    campaign.recommend()
    campaign2 = roundtrip(campaign)
    assert campaign == campaign2


def test_valid_config(config):
    Campaign.validate_config(config)


def test_invalid_config(config):
    config = config.replace("CategoricalParameter", "CatParam")
    with pytest.raises(ClassValidationError):
        Campaign.validate_config(config)



## Example for using different strategies

# This example shows how to create and use recommender objects.
# Such an object specifies the recommender adopted to make recommendations.
# It has several parameters one can adjust, depending on the recommender the user wants to follow.

# To apply the selected recommender, this object can be specified in the arguments of the campaign.
# The different parameters the user can change are:
# - The initial recommender
# - The recommender with its surrogate model and its acquisition function
# - Other parameters to allow or not repetition of recommendations


# This examples assumes some basic familiarity with using BayBE.
# We refer to [`campaign`](./campaign.md) for a more general and basic example.

### Necessary imports for this example

from baybe import Campaign
from baybe.objective import Objective
from baybe.parameters import NumericalDiscreteParameter, SubstanceParameter
from baybe.recommenders import (
    RandomRecommender,
    SequentialGreedyRecommender,
    TwoPhaseMetaRecommender,
)
from baybe.searchspace import SearchSpace
from baybe.surrogates import (
    BayesianLinearSurrogate,
    GaussianProcessSurrogate,
    NGBoostSurrogate,
    RandomForestSurrogate,
)
from baybe.targets import NumericalTarget
from baybe.utils.dataframe import add_fake_results

### Available recommenders suitable for initial recommendation

# For the first recommendation, the user can specify which recommender to use.
# The following initial recommenders are available.
# Note that it is necessary to make the corresponding import before using them.

initial_recommenders = [
    "Random",  #: RandomRecommender(),
    "Farthest Point Sampling",  # FPSRecommender(),
    "KMEANS Clustering",  # KMeansClusteringRecommender(),
]


# Per default the initial recommender chosen is a random recommender.
INITIAL_RECOMMENDER = RandomRecommender()

### Available surrogate models

# This model uses available data to model the objective function as well as the uncertainty.
# The surrogate model is then used by the acquisition function to make recommendations.

# The following are some available basic surrogates
# Use `baybe.surrogates.get_available_surrogates()` for a complete list.

available_surrogate_models = [
    GaussianProcessSurrogate(),
    RandomForestSurrogate(),
    NGBoostSurrogate(),
    BayesianLinearSurrogate(),
]

# Per default a Gaussian Process is used

SURROGATE_MODEL = GaussianProcessSurrogate()


### Acquisition function

# This function looks for points where measurements of the target value could improve the model.
# The following acquisition functions are generally available.

available_acq_functions = [
    "qPI",  # q-Probability Of Improvement
    "qEI",  # q-Expected Improvement
    "qUCB",  # q-upper confidence bound with beta of 1.0
    "PM",  # Posterior Mean,
    "PI",  # Probability Of Improvement,
    "EI",  # Expected Improvement,
    "UCB",  # upper confidence bound with beta of 1.0
]

# Note that the qvailability of the acquisition functions might depend on the `batch_size`:
#   - If `batch_size` is set to 1, all available acquisition functions can be chosen
#   - If a larger value is chosen, only those that allow batching.
#       That is, 'q'-variants of the acquisition functions must be chosen.

# The default he acquisition function is q-Expected Improvement.

ACQ_FUNCTION = "qEI"

### Other parameters

# Two other boolean hyperparameters can be specified when creating a recommender object.
# The first one allows the recommendation of points that were already recommended previously.
# The second one allows the recommendation of points that have already been measured.
# Per default, they are set to `True`.

ALLOW_REPEATED_RECOMMENDATIONS = True
ALLOW_RECOMMENDING_ALREADY_MEASURED = True

### Creating the recommender object

# To create the recommender object, each parameter described above can be specified as follows.
# Note that they all have default values.
# Therefore one does not need to specify all of them to create a recommender object.

recommender = TwoPhaseMetaRecommender(
    initial_recommender=INITIAL_RECOMMENDER,
    recommender=SequentialGreedyRecommender(
        surrogate_model=SURROGATE_MODEL,
        acquisition_function_cls=ACQ_FUNCTION,
        allow_repeated_recommendations=ALLOW_REPEATED_RECOMMENDATIONS,
        allow_recommending_already_measured=ALLOW_RECOMMENDING_ALREADY_MEASURED,
    ),
)

print(recommender)

# Note that there are the additional keywords `hybrid_sampler` and `sampling_percentag`.
# Their meaning and how to use and define it are explained in the hybrid backtesting example.
# We thus refer to [`hybrid`](./../Backtesting/hybrid.md) for details on these.

### Example Searchspace and objective parameters

# We use the same data used in the [`campaign`](./campaign.md) example.

dict_solvent = {
    "DMAc": r"CC(N(C)C)=O",
    "Butyornitrile": r"CCCC#N",
    "Butyl Ester": r"CCCCOC(C)=O",
    "p-Xylene": r"CC1=CC=C(C)C=C1",
}
dict_base = {
    "Potassium acetate": r"O=C([O-])C.[K+]",
    "Potassium pivalate": r"O=C([O-])C(C)(C)C.[K+]",
    "Cesium acetate": r"O=C([O-])C.[Cs+]",
    "Cesium pivalate": r"O=C([O-])C(C)(C)C.[Cs+]",
}
dict_ligand = {
    "BrettPhos": r"CC(C)C1=CC(C(C)C)=C(C(C(C)C)=C1)C2=C(P(C3CCCCC3)C4CCCCC4)C(OC)="
    "CC=C2OC",
    "Di-tert-butylphenylphosphine": r"CC(C)(C)P(C1=CC=CC=C1)C(C)(C)C",
    "(t-Bu)PhCPhos": r"CN(C)C1=CC=CC(N(C)C)=C1C2=CC=CC=C2P(C(C)(C)C)C3=CC=CC=C3",
}

solvent = SubstanceParameter("Solvent", data=dict_solvent, encoding="MORDRED")
base = SubstanceParameter("Base", data=dict_base, encoding="MORDRED")
ligand = SubstanceParameter("Ligand", data=dict_ligand, encoding="MORDRED")
temperature = NumericalDiscreteParameter(
    "Temperature", values=[90, 105, 120], tolerance=2
)
concentration = NumericalDiscreteParameter(
    "Concentration", values=[0.057, 0.1, 0.153], tolerance=0.005
)

# We collect all parameters in a list.

parameters = [solvent, base, ligand, temperature, concentration]

# We create the searchspace and the objective.

searchspace = SearchSpace.from_product(parameters=parameters)

objective = Objective(
    mode="SINGLE", targets=[NumericalTarget(name="yield", mode="MAX")]
)

### Creating the campaign

# The recommender object can now be used together with the searchspace and the objective as follows.

campaign = Campaign(
    searchspace=searchspace,
    recommender=recommender,
    objective=objective,
)

# This campaign can then be used to get recommendations and add measurements:

recommendation = campaign.recommend(batch_size=3)
print("\n\nRecommended experiments: ")
print(recommendation)

add_fake_results(recommendation, campaign)
print("\n\nRecommended experiments with fake measured values: ")
print(recommendation)

campaign.add_measurements(recommendation)



"""Random forest surrogates.

Currently, the documentation for this surrogate is not available. This is due to a bug
in our documentation tool, see https://github.com/sphinx-doc/sphinx/issues/11750.

Since we plan to refactor the surrogates, this part of the documentation will be
available in the future. Thus, please have a look in the source code directly.
"""

from typing import Any, ClassVar, Dict, Optional, Tuple

import numpy as np
import torch
from attr import define, field
from sklearn.ensemble import RandomForestRegressor
from torch import Tensor

from baybe.searchspace import SearchSpace
from baybe.surrogates.base import Surrogate
from baybe.surrogates.utils import batchify, catch_constant_targets, scale_model
from baybe.surrogates.validation import get_model_params_validator


@catch_constant_targets
@scale_model
@define
class RandomForestSurrogate(Surrogate):
    """A random forest surrogate model."""

    # Class variables
    joint_posterior: ClassVar[bool] = False
    # See base class.

    supports_transfer_learning: ClassVar[bool] = False
    # See base class.

    # Object variables
    model_params: Dict[str, Any] = field(
        factory=dict,
        converter=dict,
        validator=get_model_params_validator(RandomForestRegressor.__init__),
    )
    # See base class.

    _model: Optional[RandomForestRegressor] = field(init=False, default=None)
    """The actual model."""

    @batchify
    def _posterior(self, candidates: Tensor) -> Tuple[Tensor, Tensor]:
        # See base class.

        # Evaluate all trees
        # NOTE: explicit conversion to ndarray is needed due to a pytorch issue:
        # https://github.com/pytorch/pytorch/pull/51731
        # https://github.com/pytorch/pytorch/issues/13918
        predictions = torch.from_numpy(
            np.asarray(
                [
                    self._model.estimators_[tree].predict(candidates)
                    for tree in range(self._model.n_estimators)
                ]
            )
        )

        # Compute posterior mean and variance
        mean = predictions.mean(dim=0)
        var = predictions.var(dim=0)

        return mean, var

    def _fit(self, searchspace: SearchSpace, train_x: Tensor, train_y: Tensor) -> None:
        # See base class.
        self._model = RandomForestRegressor(**(self.model_params))
        self._model.fit(train_x, train_y.ravel())



"""Nonpredictive recommenders."""

from baybe.recommenders.pure.nonpredictive.clustering import (
    GaussianMixtureClusteringRecommender,
    KMeansClusteringRecommender,
    PAMClusteringRecommender,
)
from baybe.recommenders.pure.nonpredictive.sampling import (
    FPSRecommender,
    RandomRecommender,
)

__all__ = [
    "FPSRecommender",
    "GaussianMixtureClusteringRecommender",
    "KMeansClusteringRecommender",
    "PAMClusteringRecommender",
    "RandomRecommender",
]



"""Hypothesis strategies for dataframes."""

from hypothesis import strategies as st
from hypothesis.extra.pandas import column, data_frames, indexes


@st.composite
def random_dataframes(draw: st.DrawFn):
    """Generate pandas dataframes of random shape and content."""
    index_elements = st.one_of(st.text(), st.integers(), st.floats())
    cols = st.builds(
        column, name=index_elements, dtype=st.sampled_from([int, float, str])
    )
    columns = draw(st.lists(cols, max_size=5, unique_by=lambda c: c.name))
    index_strategy = indexes(elements=index_elements, max_size=2)
    return draw(data_frames(columns=columns, index=index_strategy))



"""Functionality for managing search spaces."""

from __future__ import annotations

from enum import Enum
from typing import List, Optional, cast

import pandas as pd
import torch
from attr import define, field

from baybe.constraints import (
    ContinuousLinearEqualityConstraint,
    ContinuousLinearInequalityConstraint,
    validate_constraints,
)
from baybe.constraints.base import Constraint, DiscreteConstraint
from baybe.parameters import (
    SubstanceEncoding,
    TaskParameter,
)
from baybe.parameters.base import ContinuousParameter, DiscreteParameter, Parameter
from baybe.searchspace.continuous import SubspaceContinuous
from baybe.searchspace.discrete import SubspaceDiscrete
from baybe.searchspace.validation import validate_parameters
from baybe.serialization import SerialMixin, converter, select_constructor_hook
from baybe.telemetry import TELEM_LABELS, telemetry_record_value


class SearchSpaceType(Enum):
    """Enum class for different types of search spaces and respective compatibility."""

    DISCRETE = "DISCRETE"
    """Flag for discrete search spaces resp. compatibility with discrete search
    spaces."""

    CONTINUOUS = "CONTINUOUS"
    """Flag for continuous search spaces resp. compatibility with continuous
    search spaces."""

    EITHER = "EITHER"
    """Flag compatibility with either discrete or continuous, but not hybrid
    search spaces."""

    HYBRID = "HYBRID"
    """Flag for hybrid search spaces resp. compatibility with hybrid search spaces."""


@define
class SearchSpace(SerialMixin):
    """Class for managing the overall search space.

    The search space might be purely discrete, purely continuous, or hybrid.
    Note that created objects related to the computational representations of parameters
    (e.g., parameter bounds, computational dataframes, etc.) may use a different
    parameter order than what is specified through the constructor: While the
    passed parameter list can contain parameters in arbitrary order, the
    aforementioned objects (by convention) list discrete parameters first, followed
    by continuous ones.
    """

    discrete: SubspaceDiscrete = field(factory=SubspaceDiscrete.empty)
    """The (potentially empty) discrete subspace of the overall search space."""

    continuous: SubspaceContinuous = field(factory=SubspaceContinuous.empty)
    """The (potentially empty) continuous subspace of the overall search space."""

    def __str__(self) -> str:
        start_bold = "\033[1m"
        end_bold = "\033[0m"
        print(
            f"""\n\n{start_bold}***** SEARCH SPACE *****
        \nSearch space type: {end_bold}{self.type.name}\n"""
        )

        return str(self.discrete) + str(self.continuous)

    def __attrs_post_init__(self):
        """Perform validation and record telemetry values."""
        validate_parameters(self.parameters)
        validate_constraints(self.constraints, self.parameters)

        # Telemetry
        telemetry_record_value(TELEM_LABELS["COUNT_SEARCHSPACE_CREATION"], 1)
        telemetry_record_value(TELEM_LABELS["NUM_PARAMETERS"], len(self.parameters))
        telemetry_record_value(
            TELEM_LABELS["NUM_CONSTRAINTS"],
            len(self.constraints) if self.constraints else 0,
        )

    @classmethod
    def from_product(
        cls,
        parameters: List[Parameter],
        constraints: Optional[List[Constraint]] = None,
        empty_encoding: bool = False,
    ) -> SearchSpace:
        """Create a search space from a cartesian product.

        In the search space, optional subsequent constraints are applied.
        That is, the discrete subspace becomes the (filtered) cartesian product
        containing all discrete parameter combinations while, analogously, the
        continuous subspace represents the (filtered) cartesian product of all
        continuous parameters.

        Args:
            parameters: The parameters spanning the search space.
            constraints: An optional set of constraints restricting the valid parameter
                space.
            empty_encoding: If ``True``, uses an "empty" encoding for all parameters.
                This is useful, for instance, in combination with random search
                strategies that do not read the actual parameter values, since it avoids
                the (potentially costly) transformation of the parameter values to their
                computational representation.

        Returns:
            The constructed search space.
        """
        # IMPROVE: The arguments get pre-validated here to avoid the potentially costly
        #   creation of the subspaces. Perhaps there is an elegant way to bypass the
        #   default validation in the initializer (which is required for other
        #   ways of object creation) in this particular case.
        validate_parameters(parameters)
        if constraints:
            validate_constraints(constraints, parameters)
        else:
            constraints = []

        discrete: SubspaceDiscrete = SubspaceDiscrete.from_product(
            parameters=[p for p in parameters if isinstance(p, DiscreteParameter)],
            constraints=[
                cast(DiscreteConstraint, c) for c in constraints if c.is_discrete
            ],
            empty_encoding=empty_encoding,
        )
        continuous: SubspaceContinuous = SubspaceContinuous(
            parameters=[p for p in parameters if isinstance(p, ContinuousParameter)],
            constraints_lin_eq=[
                cast(ContinuousLinearEqualityConstraint, c)
                for c in constraints
                if isinstance(c, ContinuousLinearEqualityConstraint)
            ],
            constraints_lin_ineq=[
                cast(ContinuousLinearInequalityConstraint, c)
                for c in constraints
                if isinstance(c, ContinuousLinearInequalityConstraint)
            ],
        )

        return SearchSpace(discrete=discrete, continuous=continuous)

    @classmethod
    def from_dataframe(
        cls,
        df: pd.DataFrame,
        parameters: List[Parameter],
    ) -> SearchSpace:
        """Create a search space from a specified set of parameter configurations.

        The way in which the contents of the columns are interpreted depends on the
        types of the corresponding parameter objects provided. For details, see
        :meth:`baybe.searchspace.discrete.SubspaceDiscrete.from_dataframe` and
        :meth:`baybe.searchspace.continuous.SubspaceContinuous.from_dataframe`.

        Args:
            df: A dataframe whose parameter configurations are used as
                search space specification.
            parameters: The corresponding parameter objects, one for each column
                in the provided dataframe.

        Returns:
            The created search space.

        Raises:
            ValueError: If the dataframe columns do not match with the parameters.
        """
        if set(p.name for p in parameters) != set(df.columns.values):
            raise ValueError(
                "The provided dataframe columns must match exactly with the specified "
                "parameter names."
            )

        disc_params = [p for p in parameters if isinstance(p, DiscreteParameter)]
        cont_params = [p for p in parameters if isinstance(p, ContinuousParameter)]

        return SearchSpace(
            discrete=SubspaceDiscrete.from_dataframe(
                df[[p.name for p in disc_params]], disc_params
            ),
            continuous=SubspaceContinuous.from_dataframe(
                df[[p.name for p in cont_params]], cont_params
            ),
        )

    @property
    def parameters(self) -> List[Parameter]:
        """Return the list of parameters of the search space."""
        return self.discrete.parameters + self.continuous.parameters

    @property
    def constraints(self) -> List[Constraint]:
        """Return the constraints of the search space."""
        return (
            self.discrete.constraints
            + self.continuous.constraints_lin_eq
            + self.continuous.constraints_lin_ineq
        )

    @property
    def type(self) -> SearchSpaceType:
        """Return the type of the search space."""
        if self.discrete.is_empty and not self.continuous.is_empty:
            return SearchSpaceType.CONTINUOUS
        if not self.discrete.is_empty and self.continuous.is_empty:
            return SearchSpaceType.DISCRETE
        if not self.discrete.is_empty and not self.continuous.is_empty:
            return SearchSpaceType.HYBRID
        raise RuntimeError("This line should be impossible to reach.")

    @property
    def contains_mordred(self) -> bool:
        """Indicates if any of the discrete parameters uses ``MORDRED`` encoding."""
        return any(
            p.encoding is SubstanceEncoding.MORDRED for p in self.discrete.parameters
        )

    @property
    def contains_rdkit(self) -> bool:
        """Indicates if any of the discrete parameters uses ``RDKIT`` encoding."""
        return any(
            p.encoding is SubstanceEncoding.RDKIT for p in self.discrete.parameters
        )

    @property
    def param_bounds_comp(self) -> torch.Tensor:
        """Return bounds as tensor."""
        return torch.hstack(
            [self.discrete.param_bounds_comp, self.continuous.param_bounds_comp]
        )

    @property
    def task_idx(self) -> Optional[int]:
        """The column index of the task parameter in computational representation."""
        try:
            # TODO [16932]: Redesign metadata handling
            task_param = next(
                p for p in self.parameters if isinstance(p, TaskParameter)
            )
        except StopIteration:
            return None
        # TODO[11611]: The current approach has two limitations:
        #   1.  It matches by column name and thus assumes that the parameter name
        #       is used as the column name.
        #   2.  It relies on the current implementation detail that discrete parameters
        #       appear first in the computational dataframe.
        #   --> Fix this when refactoring the data
        return self.discrete.comp_rep.columns.get_loc(task_param.name)

    @property
    def n_tasks(self) -> int:
        """The number of tasks encoded in the search space."""
        # TODO [16932]: This approach only works for a single task parameter. For
        #  multiple task parameters, we need to align what the output should even
        #  represent (e.g. number of combinatorial task combinations, number of
        #  tasks per task parameter, etc).
        try:
            task_param = next(
                p for p in self.parameters if isinstance(p, TaskParameter)
            )
            return len(task_param.values)

        # When there are no task parameters, we effectively have a single task
        except StopIteration:
            return 1

    def transform(
        self,
        data: pd.DataFrame,
    ) -> pd.DataFrame:
        """Transform data from experimental to computational representation.

        This function can e.g. be used to transform data obtained from measurements.
        Continuous parameters are not transformed but included.

        Args:
            data: The data to be transformed. Must contain all specified parameters, can
                contain more columns.

        Returns:
            A dataframe with the parameters in computational representation.
        """
        # Transform subspaces separately
        df_discrete = self.discrete.transform(data)
        df_continuous = self.continuous.transform(data)

        # Combine Subspaces
        comp_rep = pd.concat([df_discrete, df_continuous], axis=1)

        return comp_rep


def validate_searchspace_from_config(specs: dict, _) -> None:
    """Validate the search space specifications while skipping costly creation steps."""
    # For product spaces, only validate the inputs
    if specs.get("constructor", None) == "from_product":
        parameters = converter.structure(specs["parameters"], List[Parameter])
        validate_parameters(parameters)

        constraints = specs.get("constraints", None)
        if constraints:
            constraints = converter.structure(specs["constraints"], List[Constraint])
            validate_constraints(constraints, parameters)

    # For all other types, validate by construction
    else:
        converter.structure(specs, SearchSpace)


# Register deserialization hook
converter.register_structure_hook(SearchSpace, select_constructor_hook)



"""Compare different surrogate model and verify their predictions.

The purpose of this script is to compare different surrogate models and verify that
their predictions are invariant to changes in scale of the underlying target function.

This means that the shown function approximation should always appear visually the same
when the input and output scales are changed.
"""

import matplotlib.pyplot as plt
import numpy as np
import torch
from botorch.acquisition import qExpectedImprovement
from botorch.optim import optimize_acqf_discrete
from funcy import rpartial

import streamlit as st
from baybe.acquisition import debotorchize
from baybe.parameters import NumericalDiscreteParameter
from baybe.searchspace import SearchSpace
from baybe.surrogates import get_available_surrogates

# define constants
N_PARAMETER_VALUES = 1000


def cubic(
    x: np.ndarray, x_min: float, x_max: float, amplitude: float, bias: float
) -> np.ndarray:
    """Cubic test function."""
    out = amplitude * np.power((x - x_min) / (x_max - x_min), 3) + bias
    return out


def sin(
    x: np.ndarray, x_min: float, x_max: float, amplitude: float, bias: float
) -> np.ndarray:
    """Sinusoid test function."""
    out = amplitude * np.sin((x - x_min) / (x_max - x_min) * 2 * np.pi) + bias
    return out


def constant(
    x: np.ndarray, x_min: float, x_max: float, amplitude: float, bias: float
) -> np.ndarray:
    """Constant test function."""
    out = np.full(x.shape, fill_value=bias)
    return out


def linear(
    x: np.ndarray, x_min: float, x_max: float, amplitude: float, bias: float
) -> np.ndarray:
    """Linear test function."""
    out = amplitude * np.linspace(0, 1, len(x)) + bias
    return out


def main():
    """Create the streamlit dashboard."""
    # basic settings
    plt.style.use("seaborn-v0_8-paper")  # 'seaborn' is depreciated since matplotlib 3.6

    # show docstring in dashboard
    st.info(__doc__)

    # define all available test functions
    test_functions = {
        "Sine": sin,
        "Constant": constant,
        "Linear": linear,
        "Cubic": cubic,
    }

    # collect all available surrogate models
    surrogate_model_classes = {
        surr.__name__: surr for surr in get_available_surrogates()
    }

    # simulation parameters
    random_seed = int(st.sidebar.number_input("Random seed", value=1337))
    function_name = st.sidebar.selectbox("Test function", list(test_functions.keys()))
    surrogate_name = st.sidebar.selectbox(
        "Surrogate model", list(surrogate_model_classes.keys())
    )
    n_training_points = st.sidebar.slider("Number of training points", 1, 20, 5)
    n_recommendations = st.sidebar.slider("Number of recommendations", 1, 20, 5)
    st.sidebar.markdown("---")
    st.sidebar.markdown(
        """
        The plot should remain static (except for the axis labels) when changing the
        following parameters.
        """
    )
    upper_parameter_limit = st.sidebar.slider("Upper parameter limit", 0.0, 100.0, 1.0)
    lower_parameter_limit = st.sidebar.slider("Lower parameter limit", -100.0, 0.0, 0.0)
    function_amplitude = st.sidebar.slider("Function amplitude", 1.0, 100.0, 1.0)
    function_bias = st.sidebar.slider("Function bias", -100.0, 100.0, 0.0)

    # fix the chosen random seed
    np.random.seed(random_seed)
    torch.manual_seed(random_seed)

    # select the test function and the surrogate model class
    fun = rpartial(
        test_functions[function_name],
        lower_parameter_limit,
        upper_parameter_limit,
        function_amplitude,
        function_bias,
    )

    # create the input grid and corresponding target values
    test_x = torch.linspace(
        lower_parameter_limit, upper_parameter_limit, N_PARAMETER_VALUES
    )
    test_y = torch.from_numpy(fun(test_x.numpy()))

    # randomly select the specified number of training data points
    train_idx = np.random.choice(
        range(N_PARAMETER_VALUES), n_training_points, replace=False
    )
    train_x = test_x[train_idx]
    train_y = test_y[train_idx]

    # create the searchspace object
    param = NumericalDiscreteParameter(name="param", values=test_x.numpy().tolist())
    searchspace = SearchSpace.from_product(parameters=[param])

    # create the surrogate model, train it, and get its predictions
    surrogate_model = surrogate_model_classes[surrogate_name]()
    surrogate_model.fit(searchspace, train_x.unsqueeze(-1), train_y.unsqueeze(-1))

    # recommend next experiments
    # TODO: use BayBE recommender and add widgets for recommender selection
    best_f = train_y.max().item()
    acqf = debotorchize(qExpectedImprovement)(surrogate_model, best_f)
    recommendatations = optimize_acqf_discrete(
        acqf, q=n_recommendations, choices=test_x.unsqueeze(-1)
    )[0]

    # create the mean and standard deviation predictions for the entire search space
    mean, covar = surrogate_model.posterior(test_x.unsqueeze(-1))
    mean = mean.detach().numpy()
    std = covar.diag().sqrt().detach().numpy()

    # visualize the test function, training points, model predictions, recommendations
    fig = plt.figure()
    plt.plot(test_x, test_y, color="tab:blue", label="Test function")
    plt.plot(train_x, train_y, "o", color="tab:blue")
    plt.plot(test_x, mean, color="tab:red", label="Surrogate model")
    plt.fill_between(test_x, mean - std, mean + std, alpha=0.2, color="tab:red")
    plt.vlines(
        recommendatations, *plt.gca().get_ylim(), color="k", label="Recommendations"
    )
    plt.legend()
    st.pyplot(fig)


if __name__ == "__main__":
    main()



## Basic example for using BayBE

# This example shows how to create a campaign and how to use it.
# It is intended to be used as a first point of interaction with campaign after having
# read the corresponding [user guide](./../../userguide/campaigns).

### Necessary imports for this example

from baybe import Campaign
from baybe.objective import Objective
from baybe.parameters import NumericalDiscreteParameter, SubstanceParameter
from baybe.searchspace import SearchSpace
from baybe.targets import NumericalTarget
from baybe.utils.dataframe import add_fake_results

### Setup

# This example presents the optimization of a direct Arylation reaction in a discrete
# space. For this, we require data for solvents, ligands and bases.

dict_solvent = {
    "DMAc": r"CC(N(C)C)=O",
    "Butyornitrile": r"CCCC#N",
    "Butyl Ester": r"CCCCOC(C)=O",
    "p-Xylene": r"CC1=CC=C(C)C=C1",
}

dict_base = {
    "Potassium acetate": r"O=C([O-])C.[K+]",
    "Potassium pivalate": r"O=C([O-])C(C)(C)C.[K+]",
    "Cesium acetate": r"O=C([O-])C.[Cs+]",
    "Cesium pivalate": r"O=C([O-])C(C)(C)C.[Cs+]",
}

dict_ligand = {
    "BrettPhos": r"CC(C)C1=CC(C(C)C)=C(C(C(C)C)=C1)C2=C(P(C3CCCCC3)C4CCCCC4)C(OC)="
    "CC=C2OC",
    "Di-tert-butylphenylphosphine": r"CC(C)(C)P(C1=CC=CC=C1)C(C)(C)C",
    "(t-Bu)PhCPhos": r"CN(C)C1=CC=CC(N(C)C)=C1C2=CC=CC=C2P(C(C)(C)C)C3=CC=CC=C3",
}

# We define the chemical substances parameters using the dictionaries defined previously.
# Here, we use `"MORDRED"` encoding, but others are available.
# We proceed to define numerical discrete parameters `temperature` and `concentration`
# and create the search space.

solvent = SubstanceParameter("Solvent", data=dict_solvent, encoding="MORDRED")
base = SubstanceParameter("Base", data=dict_base, encoding="MORDRED")
ligand = SubstanceParameter("Ligand", data=dict_ligand, encoding="MORDRED")

temperature = NumericalDiscreteParameter(
    "Temperature", values=[90, 105, 120], tolerance=2
)
concentration = NumericalDiscreteParameter(
    "Concentration", values=[0.057, 0.1, 0.153], tolerance=0.005
)

parameters = [solvent, base, ligand, temperature, concentration]

searchspace = SearchSpace.from_product(parameters=parameters)

# In this example, we maximize the yield of a reaction and define a corresponding
# objective.

objective = Objective(
    mode="SINGLE", targets=[NumericalTarget(name="yield", mode="MAX")]
)

# We now finally create the campaign using the objects configure previously.

campaign = Campaign(
    searchspace=searchspace,
    objective=objective,
)

### Getting a recommendation and adding measurements

# We use the `recommend()` function of the campaign for getting measurements.

recommendation = campaign.recommend(batch_size=2)

print("\n\nRecommended measurements with batch_size = 2: ")
print(recommendation)

# Adding target values is done by creating a new column in the `recommendation`
# dataframe named after the target.
# In this example, we use the `add_fake_results()` utility to create fake results.
# We then update the campaign by adding the measurements.

add_fake_results(recommendation, campaign)
print("\n\nRecommended experiments with fake measured values: ")
print(recommendation)

campaign.add_measurements(recommendation)



"""Linear surrogates.

Currently, the documentation for this surrogate is not available. This is due to a bug
in our documentation tool, see https://github.com/sphinx-doc/sphinx/issues/11750.

Since we plan to refactor the surrogates, this part of the documentation will be
available in the future. Thus, please have a look in the source code directly.
"""

from typing import Any, ClassVar, Dict, Optional, Tuple

import torch
from attr import define, field
from sklearn.linear_model import ARDRegression
from torch import Tensor

from baybe.searchspace import SearchSpace
from baybe.surrogates.base import Surrogate
from baybe.surrogates.utils import batchify, catch_constant_targets, scale_model
from baybe.surrogates.validation import get_model_params_validator


@catch_constant_targets
@scale_model
@define
class BayesianLinearSurrogate(Surrogate):
    """A Bayesian linear regression surrogate model."""

    # Class variables
    joint_posterior: ClassVar[bool] = False
    # See base class.

    supports_transfer_learning: ClassVar[bool] = False
    # See base class.

    # Object variables
    model_params: Dict[str, Any] = field(
        factory=dict,
        converter=dict,
        validator=get_model_params_validator(ARDRegression.__init__),
    )
    # See base class.

    _model: Optional[ARDRegression] = field(init=False, default=None)
    """The actual model."""

    @batchify
    def _posterior(self, candidates: Tensor) -> Tuple[Tensor, Tensor]:
        # See base class.
        # Get predictions
        dists = self._model.predict(candidates.numpy(), return_std=True)

        # Split into posterior mean and variance
        mean = torch.from_numpy(dists[0])
        var = torch.from_numpy(dists[1]).pow(2)

        return mean, var

    def _fit(self, searchspace: SearchSpace, train_x: Tensor, train_y: Tensor) -> None:
        # See base class.
        self._model = ARDRegression(**(self.model_params))
        self._model.fit(train_x, train_y.ravel())



"""Validation tests for parameters."""

from unittest.mock import Mock

import numpy as np
import pandas as pd
import pytest
from attrs import NOTHING
from cattrs.errors import IterableValidationError
from pytest import param

from baybe.parameters.categorical import (
    CategoricalParameter,
    TaskParameter,
)
from baybe.parameters.custom import CustomDiscreteParameter
from baybe.parameters.numerical import (
    NumericalContinuousParameter,
    NumericalDiscreteParameter,
)
from baybe.parameters.substance import SubstanceParameter
from baybe.parameters.validation import validate_decorrelation
from baybe.utils.interval import InfiniteIntervalError

from ..conftest import _CHEM_INSTALLED

try:  # For python < 3.11, use the exceptiongroup backport
    ExceptionGroup
except NameError:
    from exceptiongroup import ExceptionGroup


@pytest.mark.parametrize(
    ("name", "error"),
    [
        param("", ValueError, id="empty_string"),
        param(1, TypeError, id="not_a_string"),
    ],
)
def test_invalid_parameter_name(name, error):
    """Providing an invalid parameter name raises an exception."""
    with pytest.raises(error):
        NumericalDiscreteParameter(name=name, values=[1, 2, 3])


@pytest.mark.parametrize(
    ("decorrelation", "error"),
    [
        param(1, TypeError, id="wrong_type"),
        param(1.0, ValueError, id="too_high"),
        param(0.0, ValueError, id="too_low"),
    ],
)
def test_invalid_decorrelation(decorrelation, error):
    """Providing an invalid decorrelation setting raises an exception."""
    with pytest.raises(error):
        validate_decorrelation(Mock(), Mock(), decorrelation)


@pytest.mark.parametrize(
    ("values", "error"),
    [
        param([1, np.inf], ValueError, id="infinity"),
        param([1, np.nan], ValueError, id="nan"),
        param([1, 1], ValueError, id="duplicates"),
        param([1], ValueError, id="only_one_value"),
        param([1, "a"], IterableValidationError, id="not_a_number"),
    ],
)
def test_invalid_values_numerical_discrete_parameter(values, error):
    """Providing invalid parameter values raises an exception."""
    with pytest.raises(error):
        NumericalDiscreteParameter(name="invalid_values", values=values)


@pytest.mark.parametrize(
    "bounds",
    [
        param([-np.inf, np.inf], id="infinite"),
        param([0, np.inf], id="open_right"),
        param([-np.inf, 0], id="open_left"),
    ],
)
def test_invalid_bounds_numerical_continuous_parameter(bounds):
    """Creating an unbounded parameter raises an exception."""
    with pytest.raises(InfiniteIntervalError):
        NumericalContinuousParameter(name="invalid_values", bounds=bounds)


@pytest.mark.parametrize(
    ("values", "error"),
    [
        param(["", "A"], ValueError, id="empty_string"),
        param(["A", "A"], ValueError, id="duplicates"),
        param(["A"], ValueError, id="only_one_value"),
        param(["A", 1], TypeError, id="not_a_string"),
    ],
)
def test_invalid_values_categorical_parameter(values, error):
    """Providing invalid parameter values raises an exception."""
    with pytest.raises(error):
        CategoricalParameter(name="invalid_values", values=values)


def test_invalid_encoding_categorical_parameter():
    """Providing an invalid encoding raises an exception."""
    with pytest.raises(ValueError):
        CategoricalParameter(name="invalid_encoding", values=["A", "B"], encoding="enc")


@pytest.mark.parametrize(
    ("values", "active_values", "error"),
    [
        param(["", "A"], NOTHING, ValueError, id="empty_string"),
        param(["A", "A"], NOTHING, ValueError, id="duplicates"),
        param(["A"], NOTHING, ValueError, id="only_one_value"),
        param(["A", 1], NOTHING, TypeError, id="not_a_string"),
        param(["A", "B"], [], ValueError, id="no_active_values"),
        param(["A", "B"], ["C"], ValueError, id="unknown_active_values"),
        param(["A", "B"], ["A", "A"], ValueError, id="duplicate_active_values"),
    ],
)
def test_invalid_values_task_parameter(values, active_values, error):
    with pytest.raises(error):
        """Providing invalid (active) parameter values raises an exception."""
        TaskParameter(name="invalid_values", values=values, active_values=active_values)


@pytest.mark.skipif(
    not _CHEM_INSTALLED, reason="Optional chem dependency not installed."
)
@pytest.mark.parametrize(
    ("data", "error"),
    [
        param({"": "C", "A": "C"}, ValueError, id="empty_string"),
        param({"A": "C"}, ValueError, id="only_one_value"),
        param({"A": "C", 1: "C"}, TypeError, id="not_a_string"),
        param({"A": "C", "B": "X", "C": "Y"}, ExceptionGroup, id="invalid_smiles"),
        param(
            {"A": "CC", "B": "C-C", "C": "CCO", "D": "OCC"},
            ExceptionGroup,
            id="duplicate_substances",
        ),
    ],
)
def test_invalid_data_substance_parameter(data, error):
    """Providing invalid substance data raises an exception."""
    with pytest.raises(error):
        SubstanceParameter(name="invalid_data", data=data)


@pytest.mark.skipif(
    not _CHEM_INSTALLED, reason="Optional chem dependency not installed."
)
def test_invalid_encoding_substance_parameter():
    """Providing an invalid encoding raises an exception."""
    with pytest.raises(ValueError):
        SubstanceParameter(
            name="invalid_encoding", data={"A": "C", "B": "CC"}, encoding="enc"
        )


@pytest.mark.parametrize(
    "data",
    [
        param(pd.DataFrame([[1, 2], [3, np.nan]], index=["A", "B"]), id="nan"),
        param(pd.DataFrame([[1, 2], [3, np.inf]], index=["A", "B"]), id="infinity"),
        param(pd.DataFrame([[1, 2], [3, 4]], index=["A", "A"]), id="duplicates"),
        param(pd.DataFrame([[1, 2]], index=["A"]), id="only_one_value"),
        param(pd.DataFrame([[1, 2], [1, 2]], index=["A", "B"]), id="zero_variance"),
        param(pd.DataFrame([[1, 2], [3, "a"]], index=["A", "B"]), id="wrong_type"),
        param(pd.DataFrame([[1, 2], [3, 4]], index=["A", 1]), id="not_a_string"),
        param(pd.DataFrame([[1, 2], [3, 4]], index=["A", ""]), id="empty_string"),
    ],
)
def test_invalid_data_custom_parameter(data):
    """Providing an invalid custom encoding raises an exception."""
    with pytest.raises(ValueError):
        CustomDiscreteParameter(name="invalid_data", data=data)



# TODO: This file needs to be refactored.
"""Tests various configurations for a small number of iterations."""

from typing import get_args, get_type_hints

import pytest

from baybe.recommenders.meta.base import MetaRecommender
from baybe.recommenders.meta.sequential import TwoPhaseMetaRecommender
from baybe.recommenders.naive import NaiveHybridSpaceRecommender
from baybe.recommenders.pure.base import PureRecommender
from baybe.recommenders.pure.bayesian.base import BayesianRecommender
from baybe.recommenders.pure.bayesian.sequential_greedy import (
    SequentialGreedyRecommender,
)
from baybe.recommenders.pure.nonpredictive.base import NonPredictiveRecommender
from baybe.searchspace import SearchSpaceType
from baybe.surrogates import get_available_surrogates
from baybe.utils.basic import get_subclasses

from .conftest import run_iterations

########################################################################################
# Settings of the individual components to be tested
########################################################################################
valid_acquisition_functions = get_args(
    get_type_hints(BayesianRecommender.__init__)["acquisition_function_cls"]
)
valid_surrogate_models = [cls() for cls in get_available_surrogates()]
valid_initial_recommenders = [cls() for cls in get_subclasses(NonPredictiveRecommender)]
# TODO the TwoPhaseMetaRecommender below can be removed if the SeqGreedy recommender
#  allows no training data
valid_discrete_recommenders = [
    TwoPhaseMetaRecommender(recommender=cls())
    for cls in get_subclasses(PureRecommender)
    if cls.compatibility
    in [SearchSpaceType.DISCRETE, SearchSpaceType.HYBRID, SearchSpaceType.EITHER]
]
# TODO the TwoPhaseMetaRecommender below can be removed if the SeqGreedy recommender
#  allows no training data
valid_continuous_recommenders = [
    TwoPhaseMetaRecommender(recommender=cls())
    for cls in get_subclasses(PureRecommender)
    if cls.compatibility
    in [SearchSpaceType.CONTINUOUS, SearchSpaceType.HYBRID, SearchSpaceType.EITHER]
]
# List of all hybrid recommenders with default attributes. Is extended with other lists
# of hybird recommenders like naive ones or recommenders not using default arguments
# TODO the TwoPhaseMetaRecommender below can be removed if the SeqGreedy recommender
#  allows no training data
valid_hybrid_recommenders = [
    TwoPhaseMetaRecommender(recommender=cls())
    for cls in get_subclasses(PureRecommender)
    if cls.compatibility == SearchSpaceType.HYBRID
]
# List of SequentialGreedy recommenders with different sampling strategies.
sampling_strategies = [
    # Valid combinations
    ("None", 0.0),
    ("None", 1.0),
    ("Farthest", 0.2),
    ("Farthest", 0.5),
    ("Random", 0.2),
    ("Random", 0.5),
]
# TODO the TwoPhaseMetaRecommender below can be removed if the SeqGreedy recommender
#  allows no training data
valid_hybrid_sequential_greedy_recommenders = [
    TwoPhaseMetaRecommender(
        recommender=SequentialGreedyRecommender(
            hybrid_sampler=sampler, sampling_percentage=per
        )
    )
    for sampler, per in sampling_strategies
]

valid_discrete_non_predictive_recommenders = [
    cls()
    for cls in get_subclasses(NonPredictiveRecommender)
    if cls.compatibility
    in [SearchSpaceType.DISCRETE, SearchSpaceType.EITHER, SearchSpaceType.HYBRID]
]
valid_discrete_bayesian_recommenders = [
    cls()
    for cls in get_subclasses(BayesianRecommender)
    if cls.compatibility
    in [SearchSpaceType.DISCRETE, SearchSpaceType.EITHER, SearchSpaceType.HYBRID]
]
# TODO the TwoPhaseMetaRecommender below can be removed if the SeqGreedy recommender
#  allows no training data
valid_naive_hybrid_recommenders = [
    TwoPhaseMetaRecommender(
        recommender=NaiveHybridSpaceRecommender(
            disc_recommender=disc, cont_recommender=SequentialGreedyRecommender()
        )
    )
    for disc in [
        *valid_discrete_non_predictive_recommenders,
        *valid_discrete_bayesian_recommenders,
    ]
]

valid_hybrid_recommenders.extend(valid_naive_hybrid_recommenders)
valid_hybrid_recommenders.extend(valid_hybrid_sequential_greedy_recommenders)

valid_meta_recommenders = get_subclasses(MetaRecommender)

test_targets = [
    ["Target_max"],
    ["Target_min"],
    ["Target_match_bell"],
    ["Target_match_triangular"],
    ["Target_max_bounded", "Target_min_bounded"],
]


# TODO: The following tests are deactivated because there is currently no Bayesian
#   recommender that can handle non-MC acquisition functions. Once the
#   MarginalRecommender (or similar) is re-added, the acqf-tests can be reactivated.
# @pytest.mark.slow
# @pytest.mark.parametrize("acquisition_function_cls", valid_acquisition_functions)
# def test_iter_acquisition_function(campaign, n_iterations, batch_size):
#     run_iterations(campaign, n_iterations, batch_size)


@pytest.mark.slow
@pytest.mark.parametrize("surrogate_model", valid_surrogate_models)
def test_iter_surrogate_model(campaign, n_iterations, batch_size):
    run_iterations(campaign, n_iterations, batch_size)


@pytest.mark.slow
@pytest.mark.parametrize("recommender", valid_initial_recommenders)
def test_iter_initial_recommender(campaign, n_iterations, batch_size):
    run_iterations(campaign, n_iterations, batch_size)


@pytest.mark.slow
@pytest.mark.parametrize("target_names", test_targets)
def test_iter_targets(campaign, n_iterations, batch_size):
    run_iterations(campaign, n_iterations, batch_size)


@pytest.mark.slow
@pytest.mark.parametrize("recommender", valid_discrete_recommenders)
def test_iter_recommender_discrete(campaign, n_iterations, batch_size):
    run_iterations(campaign, n_iterations, batch_size)


@pytest.mark.slow
@pytest.mark.parametrize("recommender", valid_continuous_recommenders)
@pytest.mark.parametrize("parameter_names", [["Conti_finite1", "Conti_finite2"]])
def test_iter_recommender_continuous(campaign, n_iterations, batch_size):
    run_iterations(campaign, n_iterations, batch_size)


@pytest.mark.slow
@pytest.mark.parametrize("recommender", valid_hybrid_recommenders)
@pytest.mark.parametrize(
    "parameter_names",
    [["Categorical_1", "SomeSetting", "Num_disc_1", "Conti_finite1", "Conti_finite2"]],
)
def test_iter_recommender_hybrid(campaign, n_iterations, batch_size):
    run_iterations(campaign, n_iterations, batch_size)


@pytest.mark.parametrize("recommender", valid_meta_recommenders, indirect=True)
def test_meta_recommenders(campaign, n_iterations, batch_size):
    run_iterations(campaign, n_iterations, batch_size)
